/*--------------------------------------------------------------
	Find line segments in document image   getlseg  v2.1
		Written   by H.Goto, Dec.  1994 (v1.0)
		Modified  by H.Goto, March 1996 (v1.6)
		Rewritten by H.Goto, Sept. 1997 (v2.0)
		Modified  by H.Goto, Oct.  1998 (v2.1)
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-1998  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


//#define		_Mon_HTS
//#define		_Mon_AllPeaks


#define		AngleReso	1024
#define		DegMul		100

#define		BWID		64
#define		CorrWid		42


			/* for Preprocessing */
//#define		RunLengthLimit	16
#define		HBlur		2

			/* for Line segment extraction */
#define		HT_Res		2
#define		HTPeakThreshold	((BWID * 3 * 1 * 3) / 4 / 4 / 4)

			/* for Analyzing first 1-dim line sample */
#define		Sample_Width0	3
#define		Sample_Width	2
#define		LenLimit	((BWID * 3) / 4)
#define		GapLimit	(BWID / 4)
#define		DCLimit		((BWID * 9 * 3) / 4 / 10)
#define		CorrMaxOffset	(BWID / 3)
#define		CorrLowerLimit	(BWID * .30)
#define		HTP_Corr_Limit	(BWID * .65)

			/* for Analyzing second 2-dim line-image sample */
#define		ImgSample_Width	16		/* 4.19mm or 11.92pt */
#define		HistWidthLimit	16		/* 2.03mm or 5.78pt */
#define		HrzHist_Bias	3
//#define		HrzHist_RBias	20
#define		HistRLS_Gap	2
#define		BLcorrlimit	(BWID * .625)
#define		BLsumlimit	80		/* in % */

			/* for Output filtering */
#define		HistWidthLimit_Del	(BWID /8)



#define	SW_MODE_ALL	0
#define	SW_MODE_RULE	1
#define	SW_MODE_TABLE	2


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<math.h>
#include	<unistd.h>

#include	<siplib.h>
#include	<sgplib.h>
#include	<ufilep.h>
#include	<imgobj.h>
#include	<objgrp.h>
#include	<Exmath.h>

#include	"peakctl.h"
#include	"gettime.h"




/*------------------------------------------------------
	Arithmetic functions
------------------------------------------------------*/

int atan_deg(int a){
	return ( (int)((DegMul * 180) * atan((double)a / (double)AngleReso) / 3.1415926536) );
}


int rotate(int a,int *x,int *y){
	double	xd,yd;
	double	an;
	xd = (double)*x;
	yd = (double)*y;
	an = atan((double)a / (double)AngleReso);
	*x = (int)(cos(an) * xd - sin(an) * yd);
	*y = (int)(sin(an) * xd + cos(an) * yd);
	return(0);
}




/*------------------------------------------------------
	Analyze Correlation Factor
------------------------------------------------------*/

int correlationi(int *d1,int *d2,int len,int *corr,int maxmove,int span){
	int	i,mv;
	int	sum;
	for ( mv=0 ; mv<maxmove ; mv++ ){
		sum = 0;
		for ( i=0 ; i<span ; i++ ){
			sum += d1[i] * d2[i + mv];
		}
		*corr++ = (BWID * sum) / CorrWid;
	}
	return(0);
}


int anl_corr(int *corr,int len,int *off,int *cf){
	int		i;
	peaklisti	*peaklst;
	int		peaks;
	int		maxpos;
	int		maxval;
	if ( 0 == (peaklst = new peaklisti[len]) )  return(-1);
	peaks = sgp_getpeaksi(corr,peaklst,len,len,0,32767);
	maxpos = 0;  maxval = 0;
	for ( i=1 ; i<peaks ; i++ ){
		if ( peaklst[i].value > maxval ){
			maxval = peaklst[i].value;
			maxpos = peaklst[i].pos;
		}
	}
	delete []peaklst;
	*off = maxpos;
	*cf  = maxval;
	return(0);
}




/*------------------------------------------------------
	Get horizontal correlation factor of Image
------------------------------------------------------*/

int get_imghrzcorr(SIPImage *image,int y1,int y2,int *corr,int maxoffset){
	int	x,y;
	int	width;
	char	*lbufc;
	int	*corrtmp;
	int	sum;
	width = image->width;
	if ( 0 == (lbufc = new char[width]) )  return(-1);
	if ( 0 == (corrtmp = new int[width]) ){
		delete []lbufc;
		return(-1);
	}
	for ( x=0 ; x<width ; x++ )  corr[x] = 0;
	sum = 0;
	for ( y=y1 ; y<=y2 ; y++ ){
		if ( image->depth == 1 )
			sip_cvt1to8(sip_getimgptr(image,y),0,lbufc,width,0,(char)0xff);
		else	memcpy(lbufc,sip_getimgptr(image,y),width);
		for ( x=0 ; x<width ; x++ ){
			if ( lbufc[x] != 0 )  ++sum;
		}
		sgp_correlationb(lbufc,lbufc,width,corrtmp,maxoffset,1,0);
//		correlationi(lbuf,lbuf,width,corrtmp,maxoffset,CorrWid);
		for ( x=0 ; x<width ; x++ )  corr[x] += corrtmp[x];
	}
	delete []corrtmp;
	delete []lbufc;
	return(sum);
}




/*------------------------------------------------------
	Least-square fitting on Line-image
------------------------------------------------------*/

int linefit(SIPImage *image,int xoff, \
			int angle,int y0,int wu,int wl,int *a,int *b){
	int	x,y,y1,xc;
	int	ys,ye;
	char	*ip,*ip0;
	int	lw;
	uchar	mask;
	ExMATH	ExMATH;
	ExMATH.lsinit(AngleReso,1,AngleReso);
	xc = (image->width +1) /2;
	ip0 = (char *)sip_getimgptr(image,0);
	lw = image->bytes_per_line;
	if ( image->depth == 1 ){
		ip0 += xoff >> 3;
		mask = 0x80;
		for ( x=0 ; x < image->width ; x++ ){
			y1 = (angle * (xc - x)) / AngleReso + y0;
			if ( 0                  >  (ys = y1 - wu) )  ys = 0;
			if ( (int)image->height <= (ye = y1 + wl) )  ye = (int)image->height ;
			ip = ip0 + (ys * lw);
			for ( y = ys ; y <= ye ; y++ ){
				if ( (*ip & mask) ){
					ExMATH._lsputdata(x - xc,y);
				}
				ip += lw;
			}
			if ( ! (mask >>= 1) ){  mask = 0x80;  ++ip0; }
		}
	}
	else{	ip0 += xoff;
		for ( x=0 ; x < image->width ; x++ ){
			y1 = (angle * (xc - x)) / AngleReso + y0;
			if ( 0      >  (ys = y1 - wu) )  ys = 0;
			if ( (int)image->height <= (ye = y1 + wl) )  ye = (int)image->height ;
			ip = ip0 + (ys * lw);
			for ( y = ys ; y <= ye ; y++ ){
				if ( ip[x] ){
					ExMATH._lsputdata(x - xc,y);
				}
				ip += lw;
			}
		}
	}
	if ( ExMATH.counts > 1 ){
		*a = -ExMATH.lsgetA();
		*b =  ExMATH.lsgetB();
	}
	else{	*a = angle;
		*b = y0;
	}
	return(0);
}




/*------------------------------------------------------
	Save image
------------------------------------------------------*/

int saveimage(SIPImage *image,char *fname){
	RASSAVE		RASSV;
	int		y;
	int		err;
	if ( RASSV.filecreat(fname,0,image->width,image->height) )  return(1);
	err = 0;
	for ( y=0 ; y < image->height ; y++ ){
		if ( RASSV.writeLine_mono((void *)sip_getimgptr(image,y),127) ){
			err = 2;
			break;
		}
	}
	if ( err ){
		RASSV.fileclose();
		return(err);
	}
	return( RASSV.fileclose() );
}


int saveimage16(SIPImage *image,char *fname){
	RASSAVE		RASSV;
	int		y,err;
	uchar		*lbuf;
	if ( 0 == (lbuf = new uchar[image->width]) )  return(1);
	if ( RASSV.filecreat(fname,1,image->width,image->height) ){
		delete []lbuf;
		return(1);
	}
	err = 0;
	for ( y=0 ; y < image->height ; y++ ){
		sip_cvt16to8u((ushort *)sip_getimgptr(image,y), \
				(uchar *)lbuf,image->width,1,1);
		RASSV.writeLine((void *)lbuf);
	}
	delete []lbuf;
	if ( err ){
		RASSV.fileclose();
		return(err);
	}
	return( RASSV.fileclose() );
}




/*------------------------------------------------------
	Load document image
------------------------------------------------------*/

SIPImage *loadimage(char *fname,int *width,int *height,int rotate,int printinfo){
	int	filetype,bufsize;
	int	y,retcode;
	int	fwidth,fheight;
	char	*lbuf;
	PBMFILE		PBMLD;
	RASLOAD		RASLD;
	IPLOAD		IPLD;
	SIPImage	*image;
	filetype = 0;
	if ( PBMLD.open(fname,"r") ){
		filetype = 1;
		if ( RASLD.fileopen(fname) ){
			filetype = 2;
			if ( IPLD.fileopen(fname) ){
				return(0);	/* File open error */
			}
		}
	}
	switch(filetype){
	    case 0:
		fwidth  = PBMLD.getwidth();
		fheight = PBMLD.getheight();
		if ( 0 != PBMLD.getfiletype() ){
			PBMLD.close();
			return(0);	/* not supported */
		}
		break;
	    case 1:
		fwidth  = RASLD.getWidth();
		fheight = RASLD.getHeight();
		if ( 0 != RASLD.getFiletype() ){
			RASLD.fileclose();
			return(0);	/* not supported */
		}
		break;
	    case 2:
		fwidth  = IPLD.getWidth();
		fheight = IPLD.getHeight();
		if ( 0 != IPLD.getFiletype() ){
			IPLD.fileclose();
			return(0);	/* not supported */
		}
		break;
	}
	if ( printinfo )  printf("Input image size : %d x %d \n",fwidth,fheight);

	bufsize = (fwidth +7) /8;
	if ( 0 == (lbuf = new char[bufsize]) )  return(0);	/* low memory */
	if ( ! rotate ){
		image = sip_CreateImage(fwidth,fheight,1);
	}
	else{	image = sip_CreateImage(fheight,fwidth,1);
	}

	retcode = 0;
	for ( y=0 ; y<fheight ; y++ ){
		switch(filetype){
		    case 0:
			retcode = PBMLD.readline(-1,(void *)lbuf);
			break;
		    case 1:
			retcode = RASLD.readLine((void *)lbuf);
			break;
		    case 2:
			retcode = IPLD.readLine((void *)lbuf);
			break;
		}
		if ( retcode )  break;
		if ( ! rotate ){
			memcpy((void *)sip_getimgptr(image,y),(void *)lbuf,bufsize);
		}
		else{	sip_PutVertLine(image,(void *)lbuf, \
				(short)(fheight -y -1),0,fwidth,1);
		}
	}

	delete []lbuf;
	switch(filetype){
	    case 0:	PBMLD.close();  break;
	    case 1:	RASLD.fileclose();  break;
	    case 2:	IPLD.fileclose();  break;
	}
	if ( retcode )  return(0);	/* read error */
	if ( ! rotate ){
		*width = fwidth;  *height = fheight;
	}
	else{	*width = fheight;  *height = fwidth;
	}
	return(image);
}




/*------------------------------------------------------
	Filter out Long-Run
------------------------------------------------------*/

int filter_LongRun(CRects *CR,int length){
	int	crn;
	int	acr;
	for ( acr=crn=1 ; crn < CR->counts ; crn++ ){
		if ( ! CR->isactive(crn) )  continue;
		if ( CR->height(crn) > length ){
			CR->destroy(crn);
			continue;
		}
		++acr;
	}
	return(acr);
}




/*------------------------------------------------------
	Hough Transform
------------------------------------------------------*/

int HT_DrawLine(SIPImage *image,int x,int y,int hw,int weight){
	int	a,b;
	int	yoff;
	int	bw,as,ae;
	ushort	*hp;
	bw = image->width /2;
	as = -bw;
	ae = bw -1;
	if ( hw != -1 ){
		if ( as < -hw )  as = -hw;
		if ( ae > hw  )  ae = hw;
	}
	yoff = (image->height -1) /2 -y;
	hp = (ushort *)sip_getimgptr(image,0);
	if ( weight == 1 ){
		for ( a = as ; a <= ae ; a++ ){
			b = yoff + (x * a) / bw;
			++hp[ (a + bw) + ((image->bytes_per_line * b) / sizeof(ushort)) ];
		}
	}
	else{	for ( a = as ; a <= ae ; a++ ){
			b = yoff + (x * a) / bw;
			hp[ (a + bw) + ((image->bytes_per_line * b) / sizeof(ushort)) ] += weight;
		}
	}
	return(0);
}


int HT_Draw(SIPImage *image,CRects *CR,int xoff,int yoff,int hw){
	int	crn;
	CRect	*rect;
	sip_ClearImage(image,0);
	for ( crn=0 ; crn < CR->counts ; crn++ ){
		if ( ! CR->isactive(crn) )  continue;
		rect = CR->getrect(crn);
#ifndef HBlur
		HT_DrawLine(image,(rect->x1 - xoff) /HT_Res, \
				(yoff - ((rect->y2 + rect->y1) /2)) /HT_Res ,hw,1);
#else
		HT_DrawLine(image,(rect->x1 - xoff) /HT_Res, \
				(yoff - ((rect->y2 + rect->y1) /2)) /HT_Res ,hw,HBlur);
#endif
	}
	return(0);
}




/*------------------------------------------------------
	Create RL's
------------------------------------------------------*/

int create_RLs(SIPImage *image,int xoff,int width,CRects *CR,CRects *BCR){
	int	x,y,retcode;
	int	*lc,*lc2;
	char	*lbuf,*lbuf2;
	int	height;
	height = image->height;
	if ( 0 == (lc = new int[2 * width]) )  return(-1);
	lc2 = lc + width;
	if ( 0 == (lbuf = new char[2 * width]) ){
		delete []lc;
		return(-1);
	}
	lbuf2 = lbuf + width;
	for ( x=0 ; x<width ; x++ )  lc[x] = lc2[x] = 0;

	retcode = 0;
	CR->clear();
#ifdef HBlur
	BCR->clear();
#endif
	for ( y=0 ; (y < height) && (retcode == 0) ; y++ ){
		sip_cvt1to8((char *)sip_getimgptr(image,y) + (xoff >> 3), \
							0,lbuf,width,(char)0xff,0);
#ifdef HBlur
		for ( x=0 ; x<width ; x+=HBlur ){
#if HBlur == 3
			if ( lbuf[x] | lbuf[x+1] | lbuf[x+2] ){
#else
			if ( lbuf[x] | lbuf[x+1] ){
#endif
				++lc2[x];
			}
			else{	if ( lc2[x] != 0 ){
					if ( 0 > BCR->create(x,y - lc2[x],x,y - 1) ){
						retcode = -1;  break;
					}
					lc2[x] = 0;
				}
			}
		}
		if ( retcode )  break;
#endif
		for ( x=0 ; x<width ; x++ ){
			if ( lbuf[x] ){
				++lc[x];
			}
			else{	if ( lc[x] != 0 ){
					if ( 0 > CR->create(x,y - lc[x],x,y - 1) ){
						retcode = -1;  break;
					}
					lc[x] = 0;
				}
			}
		}
	}
	for ( x=0 ; (x < width) && (retcode == 0) ; x++ ){
		if ( lc[x] != 0 ){
			if ( CR->create(x,y - lc[x],x,y - lc[x]) )  retcode = -1;
		}
	}
#ifdef HBlur
	for ( x=0 ; (x < width) && (retcode == 0) ; x+=HBlur ){
		if ( lc2[x] != 0 ){
			if ( BCR->create(x,y - lc2[x],x,y - lc2[x]) )  retcode = -1;
		}
	}
#endif

	delete []lc;
	delete []lbuf;
	return(0);
}


int create_DOTs(SIPImage *image,int xoff,int width,CRects *CR,CRects *BCR){
	int	x,y,retcode;
	char	*lbuf;
	char	*lp;
	int	height;
	height = image->height;
	if ( 0 == (lbuf = new char[width]) )  return(-1);

	retcode = 0;
	CR->clear();
#ifdef HBlur
	BCR->clear();
#endif
	if ( image->depth == 1 ){
		for ( y=0 ; (y < height) && (retcode == 0) ; y++ ){
			sip_cvt1to8((char *)sip_getimgptr(image,y) + (xoff >> 3), \
								0,lbuf,width,(char)0xff,0);
#ifdef HBlur
			for ( x=0 ; x<width ; x+=HBlur ){
#if HBlur == 3
				if ( lbuf[x] | lbuf[x+1] | lbuf[x+2] ){
#else
				if ( lbuf[x] | lbuf[x+1] ){
#endif
					if ( 0 > BCR->create(x,y,x,y) ){
							retcode = -1;  break;
					}
				}
			}
			if ( retcode )  break;
#endif
			for ( x=0 ; x<width ; x++ ){
				if ( lbuf[x] ){
					if ( 0 > CR->create(x,y,x,y) ){
						retcode = -1;  break;
					}
				}
			}
		}
	}
	else{	for ( y=0 ; (y < height) && (retcode == 0) ; y++ ){
			lp = (char *)sip_getimgptr(image,y) + xoff;
#ifdef HBlur
			for ( x=0 ; x<width ; x+=HBlur ){
#if HBlur == 3
				if ( 0 == (lp[x] & lp[x+1] & lp[x+2]) ){
#else
				if ( 0 == (lp[x] & lp[x+1]) ){
#endif
					if ( 0 > BCR->create(x,y,x,y) ){
							retcode = -1;  break;
					}
				}
			}
			if ( retcode )  break;
#endif
			for ( x=0 ; x<width ; x++ ){
				if ( 0 == lp[x] ){
					if ( 0 > CR->create(x,y,x,y) ){
						retcode = -1;  break;
					}
				}
			}
		}
	}

	delete []lbuf;
	return(0);
}




/*------------------------------------------------------
	Line segment list control
------------------------------------------------------*/

#define		LSEGSTAT_ACTIVE	0x8000

typedef struct {
	int	stat;
	int	attr;
	int	x,y;
	int	a;
	int	htpeak;
	int	corroff;
	int	corrpeak;
	int	wu,wl;
} lsegpac;



class LSEGS {
    private:
	lsegpac*	lseglist;
	int		maxlsegs;
	int		incalloc();
	int		out_count;
    public:
	int		counts;

	int		clear();
	int		trunc();
	int		setnewlseg(int x,int y,int a,int htpeak, \
				int corroff,int corrpeak,int wu,int wl);
	int		deletelseg(int pkn);
	int		isactive(int pkn){ \
				return ( lseglist[pkn].stat & LSEGSTAT_ACTIVE ); }
	int		setattr(int pkn,int attr);
	int		getattr(int pkn);
	int		setfirst(){  out_count = 0; return 0; }
	int		getnext(int *id,int *x,int *y,int *a,int *htpeak, \
				int *corroff,int *corrpeak,int *wu,int *wl);
	lsegpac*	getlseg(int lsn);
	int		sortbyoffset();
			LSEGS();
			~LSEGS();
};


LSEGS :: LSEGS(){
	lseglist = 0;
	maxlsegs = 0;
	counts = 0;
}

LSEGS :: ~LSEGS(){
	clear();
}


int LSEGS :: incalloc(){
	lsegpac*	pktmp;
	int		newsize;
	newsize = maxlsegs + 256;
	if ( 0 == (pktmp = new lsegpac[newsize]) )  return(-1);
	memcpy((void *)pktmp,(void *)lseglist,maxlsegs * sizeof(lsegpac));
	delete []lseglist;
	lseglist = pktmp;
	maxlsegs = newsize;
	return(0);
}


int LSEGS :: clear(){
	maxlsegs = counts = 0;
	if ( lseglist != 0 ){  delete []lseglist;  lseglist = 0; }
	return(0);
}


int LSEGS :: trunc(){
	int		pkn,newcount;
	lsegpac*	pktmp;
	pktmp = lseglist;
	newcount = 0;
	for ( pkn=0 ; pkn<counts ; pkn++ ){
		if ( ! isactive(pkn) )  continue;
		++newcount;
		*pktmp++ = lseglist[pkn];
	}
	counts = newcount;
	return(0);
}


int LSEGS :: setnewlseg(int x,int y,int a,int htpeak, \
			int corroff,int corrpeak,int wu,int wl){
	if ( counts >= maxlsegs ){
		if ( incalloc() )  return(-1);
	}
	lseglist[counts].stat = LSEGSTAT_ACTIVE;
	lseglist[counts].attr = 0;
	lseglist[counts].x = x;
	lseglist[counts].y = y;
	lseglist[counts].a = a;
	lseglist[counts].htpeak = htpeak;
	lseglist[counts].corroff  = corroff;
	lseglist[counts].corrpeak = corrpeak;
	lseglist[counts].wu = wu;
	lseglist[counts].wl = wl;
	++counts;
	return ( counts -1 );
}


int LSEGS :: deletelseg(int pkn){
	if ( pkn >= maxlsegs )  return(-1);
	lseglist[pkn].stat = 0;
	return(0);
}


int LSEGS :: setattr(int pkn,int attr){
	if ( pkn >= maxlsegs )  return(-1);
	lseglist[pkn].attr = attr;
	return(0);
}


int LSEGS :: getattr(int pkn){
	if ( pkn >= maxlsegs )  return(0);
	return( lseglist[pkn].attr );
}


int LSEGS :: getnext(int *id,int *x,int *y,int *a,int *htpeak, \
			int *corroff,int *corrpeak,int *wu,int *wl){
	for ( ; out_count < counts ; out_count++ ){
		if ( isactive(out_count) ){
			if ( id != 0 )  *id = out_count;
			*x = lseglist[out_count].x;
			*y = lseglist[out_count].y;
			*a = lseglist[out_count].a;
			*htpeak = lseglist[out_count].htpeak;
			*corroff = lseglist[out_count].corroff;
			*corrpeak = lseglist[out_count].corrpeak;
			if ( wu != 0 )  *wu = lseglist[out_count].wu;
			if ( wl != 0 )  *wl = lseglist[out_count].wl;
			++out_count;
			return(0);
		}
	}
	return(-1);	/* End of list */
}


lsegpac* LSEGS :: getlseg(int lsn){
	if ( lsn >= maxlsegs )  return(0);
	return ( &lseglist[lsn] );
}


/*---------------------------- This function is not included in Class. */
int lsegs_cmp_offset(const void *c1,const void *c2){
	return ( ((lsegpac *)c1)->y - ((lsegpac *)c2)->y );
}
/*---------------------------------------------------------------------*/


int LSEGS :: sortbyoffset(){
	if ( trunc() )  return(-1);
	qsort((char *)lseglist,counts,sizeof(lsegpac),lsegs_cmp_offset);
	return(0);
}




/*------------------------------------------------------
	Line-segment Pool class
------------------------------------------------------*/

class LSEGPOOL : public D8SAVE,public LSEGS {
    private:
    public:
	int		savelist(char *fname,int width,int height);
};


int LSEGPOOL :: savelist(char *fname,int width,int height){
	int	i;
	d8file	d8h;
	int	x,y,a,htp,off,cf,wu,wl;
	struct {
		short	x,y;
		short	a;
		short	htpeak;
		short	corroff,corrpeak;
		short	wu,wl;
	} d8data;
	for ( i=0 ; i<8 ; i++ )  d8h.d8_llimit[i] = d8h.d8_hlimit[i] = 0;
	d8h.d8_hlimit[0] = width -1;
	d8h.d8_hlimit[1] = height -1;
	d8h.d8_llimit[2] = -AngleReso;
	d8h.d8_hlimit[2] = AngleReso;
	d8h.d8_hlimit[3] = 32767;
	d8h.d8_llimit[4] = -1;
	d8h.d8_hlimit[4] = BWID;
	d8h.d8_hlimit[5] = 32767;
	d8h.d8_hlimit[6] = ImgSample_Width +1;
	d8h.d8_hlimit[7] = ImgSample_Width +1;
	if ( creat(fname,&d8h) )  return(-1);
	setfirst();
	while(1){
		if ( -1 == getnext(0,&x,&y,&a,&htp,&off,&cf,&wu,&wl) )  break;
		d8data.x = (short)x;
		d8data.y = (short)y;
		d8data.a = (short)a;
		d8data.htpeak = (short)htp;
		d8data.corroff = (short)off;
		d8data.corrpeak = (short)cf;
		d8data.wu = (short)wu;
		d8data.wl = (short)wl;
		if ( putdata((d8pac *)&d8data) ){
			close();
			unlink(fname);
			return(-2);
		}
	}
	if ( close() )  return(-1);
	return(0);
}




/*------------------------------------------------------
	Miscellaneous rules
------------------------------------------------------*/

int check_vrthist(int wu,int wl){
	if ( (wu + wl + 1) > HistWidthLimit )  return(-1);	/* Broad  */
	return(0);						/* Narrow */
}


int check_vrthist_del(int wu,int wl){
	if ( (wu + wl + 1) > HistWidthLimit_Del )  return(-1);	/* Broad  */
	return(0);						/* Narrow */
}


int check_htcorr(int htp,int cf){
	if ( (htp + cf) < HTP_Corr_Limit )  return(0);		/* False */
	return(-1);						/* True  */
}


int check_BLcorrsum(int cf,int sum){
	if ( cf >= BLcorrlimit )  return(-1);			/* BL enable  */
	if ( sum >= BLsumlimit )  return(-1);			/* BL enable  */
	return(0);						/* BL disable */
}




/*------------------------------------------------------
	Band operation class
------------------------------------------------------*/


class BandOp : public PeakSuppress {
    private:
	int		width,height;
	int*		corr;
	int*		imgcorr;
	SIPImage*	lineb;
	SIPImage*	line;
	SIPImage*	hsp;
	SIPImage*	dotimage;
	SIPImage*	imgsamp2;
	class CRects	CRects;
	class CRects	CRectsB;
	peaklisti	*peaklst;
	PEAKS		PEAKPOOL;
	LSEGS		LSEGLOCAL;

	void		putVLine(int x,int y0,int len,int *buf);
	void		getVLine(int x,int y0,int len,int *buf);

	int		makedotimage();
	int		merge_peaks(int mul,int htres);

	int		getattnline(SIPImage *image,int xoff,int angle,int y0,int w, \
						char *lbuf,SIPImage *imgbuf);
	int		get_vrtwidth(SIPImage *image,int xoff, \
					int a,int b,int bias,int *wu,int *wl);

	void		printlbuf(char *lbuf,int len);
	int		mon_saveimage(int wu,int wl);

    public:
	double		time_rlencode,	time_hough,	time_peakdet;
	double		time_lsegcomp,	time_linecorr,	time_areacorr;

	int		init(int width,int height);
	int		clear();

	int		ope(SIPImage *image,SIPImage *thinimage,int xoff, \
			  LSEGS *LSEGLST,int mode,int infosw,int directsw,int hw);

			BandOp();
	virtual		~BandOp();
};


BandOp :: BandOp(){
	corr = imgcorr = 0;
	hsp = lineb = line = dotimage = imgsamp2 = 0;
	peaklst = 0;
}


BandOp :: ~BandOp(){
	clear();
}


int BandOp :: clear(){
	if ( corr != 0 ){  delete []corr;  corr = 0; }
	if ( imgcorr != 0 ){  delete []imgcorr;  imgcorr = 0; }
	if ( lineb != 0 ){  sip_DestroyImage(lineb);  lineb = 0;  }
	if ( line != 0 ){  sip_DestroyImage(line);  line = 0;  }
	if ( hsp != 0 ){  sip_DestroyImage(hsp);  hsp = 0;  }
	if ( dotimage != 0 ){  sip_DestroyImage(dotimage);  dotimage = 0;  }
	if ( imgsamp2 != 0 ){  sip_DestroyImage(imgsamp2);  imgsamp2 = 0;  }
	if ( peaklst != 0 ){  delete []peaklst;  peaklst = 0; }
	return(0);
}


int BandOp :: init(int width,int height){
	int	hheight;
	clear();
	BandOp :: width  = width;
	BandOp :: height = height;
	hheight = (width + height +1) & ~0x01;
	corr = new int[width];
	imgcorr = new int[width];
	lineb = sip_CreateImage(width,1,8);
	line = sip_CreateImage(hheight,1,8 * sizeof(int));
	hsp = sip_CreateImage(width / HT_Res,hheight / HT_Res,16);
//	dotimage = sip_CreateImage(width,height,8);
	dotimage = sip_CreateImage(width,height,1);
	imgsamp2 = sip_CreateImage(width,2 * ImgSample_Width +1,8);
	peaklst = new peaklisti[hheight];
	if ( (corr == 0) || (imgcorr == 0) || (imgsamp2 == 0) \
		|| (lineb == 0) || (line == 0) || (dotimage == 0) \
		|| (hsp == 0) || (peaklst == 0) ){
		clear();
		return(-1);
	}
	if ( allocpeakbuf(hheight / HT_Res) ){
		clear();
		return(-1);
	}
	return(0);
}


void BandOp :: putVLine(int x,int y0,int len,int *buf){
	ushort	*hp;
	int	lw;
	hp = (ushort *)sip_getimgptr(hsp,y0) + x;
	lw = hsp->bytes_per_line / sizeof(ushort);
	for ( ; len > 0 ; len--, hp += lw )  *hp = (ushort)*buf++;
}


void BandOp :: getVLine(int x,int y0,int len,int *buf){
	ushort	*hp;
	int	lw;
	hp = (ushort *)sip_getimgptr(hsp,y0) + x;
	lw = hsp->bytes_per_line / sizeof(ushort);
	for ( ; len > 0 ; len--, hp += lw )  *buf++ = (int)*hp;
}


int BandOp :: makedotimage(){
	int	x,y;
	int	i;
	CRect	*rect;
	char	*ip;
	int	lw;
	sip_ClearImage(dotimage,0);
	ip = (char *)sip_getimgptr(dotimage,0);
	lw = dotimage->bytes_per_line;
	for ( i=0 ; i < CRects.counts ; i++ ){
		if ( ! CRects.isactive(i) )  continue;
		rect = CRects.getrect(i);
		x = rect->x1;
		y = (rect->y2 + rect->y1) /2;
		if ( dotimage->depth == 1 ){
			(ip + (y * lw))[x>>3] |= (uchar)(0x80 >> (x & 0x07));
		}
		else{	(ip + (y * lw))[x] = (char)0xff;
		}
	}
	return(0);
}


int BandOp :: getattnline(SIPImage *image,int xoff,int angle, \
			int y0,int w,char *lbuf,SIPImage *imgbuf){
	int	x,y,yt,y1,xc;
	int	ys,ye;
	char	*ip;
	char	*lp;
	int	lw;
	uchar	mask;
	xc = (width +1) /2;
	ip = (char *)sip_getimgptr(image,0);
	lw = image->bytes_per_line;
	if ( lbuf != 0 ){
		for ( x=0 ; x < width ; x++ )  lbuf[x] = 0;
	}
	if ( imgbuf != 0 )  sip_ClearImage(imgbuf,0);
	if ( image->depth == 1 ){
		ip += xoff >> 3;
		mask = 0x80;
		if ( imgbuf == 0 ){
			for ( x=0 ; x < width ; x++ ){
				y1 = (angle * (xc - x)) / AngleReso + y0;
				if ( 0      >  (ys = y1 - w) )  ys = 0;
				if ( height <= (ye = y1 + w) )  ye = height ;
				for ( y = ys ; y <= ye ; y++ ){
					if ( (*(ip + (y * lw)) & mask) ){
						if ( lbuf != 0 )  lbuf[x] |= (char)0xff;
					}
				}
				if ( ! (mask >>= 1) ){  mask = 0x80;  ++ip; }
			}
		}
		else{	for ( x=0 ; x < width ; x++ ){
				lp = (char *)sip_getimgptr(imgbuf,0) + x;
				y1 = (angle * (xc - x)) / AngleReso + y0;
				if ( 0 >  (ys = y1 - w) ){
					lp += (imgbuf->bytes_per_line * (-ys));
					ys = 0;
				}
				if ( height <= (ye = y1 + w) )  ye = height ;
				for ( y = ys ; y <= ye ; y++ ){
					if ( (*(ip + (y * lw)) & mask) ){
						if ( lbuf != 0 )  lbuf[x] |= (char)0xff;
						*lp = (char)0xff;
					}
					lp += imgbuf->bytes_per_line;
				}
				if ( ! (mask >>= 1) ){  mask = 0x80;  ++ip; }
			}
		}
	}
	else{	ip += xoff;
		if ( imgbuf == 0 ){
			for ( x=0 ; x < width ; x++ ){
				y1 = (angle * (xc - x)) / AngleReso + y0;
				for ( yt = -w ; yt <= w ; yt++ ){
					y = y1 + yt;
					if ( (y < 0) || (y >= height) )  continue;
					if ( (ip + (y * lw))[x] ){
						if ( lbuf != 0 )  lbuf[x] |= (char)0xff;
					}
				}
			}
		}
		else{	for ( x=0 ; x < width ; x++ ){
				lp = (char *)sip_getimgptr(imgbuf,0) + x;
				y1 = (angle * (xc - x)) / AngleReso + y0;
				for ( yt = -w ; yt <= w ; yt++ ){
					y = y1 + yt;
					if ( (y >= 0) && (y < height) ){
						if ( (ip + (y * lw))[x] ){
							if ( lbuf != 0 )  lbuf[x] |= (char)0xff;
							*lp = (char)0xff;
						}
					}
					lp += imgbuf->bytes_per_line;
				}
			}
		}
	}
	return(0);
}


void BandOp :: printlbuf(char *lbuf,int len){
	int	i;
	for ( i=0 ; i<len ; i++ ){
		if ( lbuf[i] )  printf("*");  else  printf("-");
	}
	printf("\n");
}


int BandOp :: get_vrtwidth(SIPImage *image,int xoff,int a,int b, \
					int bias,int *wu,int *wl){
	int	ht_ares;
	int	*rden,sum;
	int	x,y;
	char	*lp;
	int	u,l;
#ifdef HrzHist_RBias
	int	hmax;
#endif
	if ( 0 == (rden = new int[width]) )  return(-1);
	ht_ares = hsp->width /2;
	getattnline(image,xoff,a,b,ImgSample_Width,0,imgsamp2);
	for ( y=0 ; y<(2 * ImgSample_Width +1) ; y++ ){
		lp = (char *)sip_getimgptr(imgsamp2,y);
		sum = 0;
		for ( x=0 ; x<width ; x++ ){
			if ( lp[x] != 0 )  ++sum;
		}
		rden[y] = sum;
	}

#ifdef HrzHist_RBias
	if ( bias == -1 ){	/* Auto thresholding mode */
		hmax = 0;
		for ( y = -2 ; y <= 2 ; y++ ){
			sum = rden[ ImgSample_Width + y ];
			if ( hmax < sum )  hmax = sum;
		}
		bias = (hmax * HrzHist_RBias) /100;
	}
#endif

#if HistRLS_Gap > 0
	for ( y=l=0 ; y<(2 * ImgSample_Width +1) ; y++ ){
		if ( (rden[y] - bias) > 0 ){
			if ( l != 0 ){
				if ( l <= HistRLS_Gap ){
					for ( x=u ; x<y ; x++ )  rden[x] = bias +1;
				}
				l = 0;
			}
		}
		else{	if ( l == 0 )  u = y;
			++l;
		}
	}
#endif

	u = l = 0;
	for ( y=ImgSample_Width -1 ; y>=0 ; y-- ){
		if ( (rden[y] - bias) <= 0 )  break;
		++u;
	}
	for ( y=ImgSample_Width +1 ; y<(2 * ImgSample_Width +1) ; y++ ){
		if ( (rden[y] - bias) <= 0 )  break;
		++l;
	}

	delete []rden;
	*wu = u;
	*wl = l;
	return(0);
}


static int	save_flag = 0;
static int	fnn = 0;
static char	fname_attn0[] = "attn/";
static char	fname_attn[16];


int BandOp :: mon_saveimage(int wu,int wl){
	int	x,y;
	char	*p;
	if ( ! save_flag )  return(0);
	for ( y=0 ; y < (ImgSample_Width - wu) ; y++ ){
		for ( x=0 ; x < imgsamp2->width ; x++ ){
			p = (char *)sip_getimgptr(imgsamp2,y)[x];
			*p = 0;
		}
	}
	for ( y=(ImgSample_Width + wl) +1 ; y < (2 * ImgSample_Width +1) ; y++ ){
		for ( x=0 ; x < imgsamp2->width ; x++ ){
			p = (char *)sip_getimgptr(imgsamp2,y)[x];
			*p = 0;
		}
	}
	sprintf(fname_attn,"%s%04d",fname_attn0,fnn++);
	if ( saveimage(imgsamp2,fname_attn) )  save_flag = 0;
	return(0);
}


int BandOp :: merge_peaks(int mul,int htres){
	int		s1,s2;
	peakpac		*peak1,*peak2;
	int		vr,vm;
	for ( s2=0 ; s2<PEAKPOOL.counts ; s2++ ){
		if ( ! PEAKPOOL.isactive(s2) )  continue;
		peak2 = PEAKPOOL.getpeak(s2);
		for ( s1=(s2+1) ; s1<PEAKPOOL.counts ; s1++ ){
			if ( ! PEAKPOOL.isactive(s1) )  continue;
			peak1 = PEAKPOOL.getpeak(s1);
			if ( abs(htres * (peak1->y - peak2->y)) > 2 )  continue;
			if ( (abs(peak1->x - peak2->x) / mul) > 3 )  continue;
			if ( peak1->value < peak2->value ){
				vr = 100 * peak1->value / peak2->value;
				vm = peak2->value;
			}
			else{	vr = 100 * peak2->value / peak1->value;
				vm = peak1->value;
			}
			if ( vr >= 70 ){
				peak1->x = peak2->x = (peak1->x + peak2->x) /2;
				peak1->y = peak2->y = (peak1->y + peak2->y) /2;
				peak1->value = peak2->value = vm;
			}
			if ( peak1->value < peak2->value ){
				PEAKPOOL.deletepeak(s1);
			}
			else{	PEAKPOOL.deletepeak(s2);
				break;
			}
		}
	}
	return(0);
}


int BandOp :: ope(SIPImage *image,SIPImage *thinimage, \
	int xoff,LSEGS *LSEGLST,int mode,int infosw,int directsw,int hw){
	int		x;
	int		errflag;
	int		value;
	int		xctr;
	int		ht_ares,bw;
	char*		lbufb;
	int*		lbuf;
	int		peaks,allpeaks;
	int		fmask[] = {-2,1,2,1,-2};
	int		fmask_corr[] = {1,4,1};
	lbufb = (char *)sip_getimgptr(lineb,0);
	lbuf  = (int *)sip_getimgptr(line,0);
	ht_ares = hsp->width /2;
	xctr = xoff + (width /2);

	time_start();  time_rlencode = 0;
	if ( thinimage != 0 ){
		if ( create_DOTs(thinimage,xoff,width,&CRects,&CRectsB) )  return(1);
	}
	else if ( directsw ){
		if ( create_DOTs(image,xoff,width,&CRects,&CRectsB) )  return(1);
	}
	else{	if ( create_RLs(image,xoff,width,&CRects,&CRectsB) )  return(1);
						/* Too many rectangles. */
	}
	time_get(&time_rlencode,0);

	if ( infosw ){
		printf("--------\n");
		printf("%d rectangles found. \n",CRects.counts -1);
#ifdef RunLengthLimit
		printf("%d active rectangles. \n",filter_LongRun(&CRects,RunLengthLimit));
#endif
	}


	time_start();  time_hough = 0;
#ifdef HBlur
	HT_Draw(hsp,&CRectsB,width /2,(height -1) /2 ,hw);
#else
	HT_Draw(hsp,&CRects,width /2,(height -1) /2 ,hw);
#endif
	time_get(&time_hough,0);
	if ( infosw )  printf("Hough Transform ... done.\n");


	clearpeakbuf();
	allpeaks = 0;
	PEAKPOOL.clear();
	LSEGLOCAL.clear();
	time_start();  time_peakdet = 0;
	errflag = 0;
	bw = hsp->width /2;
	for ( x=0 ; x < hsp->width ; x++ ){
		if ( hw != -1 ){
			if ( (x - bw) < -hw )  continue;
			if ( (x - bw) >  hw )  continue;
		}
		getVLine(x,0,hsp->height,lbuf);
		sgp_convolutei(lbuf,lbuf,hsp->height,0,fmask,2,5);
//		clipint(lbuf,lbuf,hsp->height,0,32767);
		peaks = sgp_getpeaksi(lbuf,peaklst, \
			hsp->height,hsp->height,HTPeakThreshold,32767);
		allpeaks += peaks;
		if ( setpeaks(x,peaklst,peaks,&PEAKPOOL,2) ){
			errflag = 1;  break;
		}
	}
	if ( setpeaks(x,peaklst,0,&PEAKPOOL,2) )  errflag = 1;
	if ( infosw )  printf("Total %d peaks.\n",allpeaks);
	if ( errflag ){
		printf("Error: Peak-local over flow.\n");
		return(2);
	}
	time_get(&time_peakdet,0);
	if ( infosw )  printf("Peak detection ... done.\n");

#ifdef	_Mon_HTS
	if ( saveimage16(hsp,"tmp.sun") )  return(1);
#endif


	time_start();  time_lsegcomp = 0;
	makedotimage();
	merge_peaks(1,HT_Res);
	PEAKPOOL.sortbyvalue();
	PEAKPOOL.setfirst();
	while (1){
		int	a,b;
		int	id;
		if ( -1 == PEAKPOOL.getnext(&id,&a,&b,&value) )  break;
		a -= ht_ares;
		a = (AngleReso * a) / ht_ares;
		b = ((b - (hsp->height -1) /2) * HT_Res) + ((height -1) /2);
		linefit(dotimage,0,a,b,Sample_Width0,Sample_Width0,&a,&b);
		PEAKPOOL.getpeak(id)->x = a;
		PEAKPOOL.getpeak(id)->y = b;
	}
	merge_peaks(AngleReso/ht_ares,1);
	time_get(&time_lsegcomp,0);
	PEAKPOOL.setfirst();
	time_start();  time_linecorr = 0;
	while (1){
		int	pkx,pky;
		int	off,cf;
		int	sp,ep,len;
		int	gl,gap,dc;
		if ( -1 == PEAKPOOL.getnext(0,&pkx,&pky,&value) )  break;
		getattnline(dotimage,0,pkx,pky,Sample_Width,lbufb,0);
		dc = 0;
		sp = -1;  ep = 0;
		gap = gl = 0;
		for ( x=0 ; x<width ; x++ ){
			if ( lbufb[x] ){
				lbuf[x] = 1;  ++dc;
				if ( sp == -1 )  sp = x;
				ep = x;
				if ( gl != 0 ){
					if ( gap < gl )  gap = gl;
					gl = 0;
				}
			}
			else{	lbuf[x] = -1;
				++gl;
			}
		}
		if ( (gl != 0) && (gap < gl) )  gap = gl;
		len = ep - sp + 1;
		off = 0;
		if ( dc >= DCLimit ){
			off = -1;
			cf = -dc;
		}
		else{
			sgp_correlationb(lbufb,lbufb,width,corr,CorrMaxOffset,1,0);
//			correlationi(lbuf,lbuf,width,corr,CorrMaxOffset,CorrWid);
			sgp_convolutei(corr,corr,CorrMaxOffset,0,fmask_corr,6,3);
			anl_corr(corr,CorrMaxOffset,&off,&cf);
		}
		if ( (off != 0) && (len >= LenLimit) && (gap <= GapLimit) ){
			if ( abs(cf) >= CorrLowerLimit ){
#ifdef _Mon_AllPeaks
				printf("Peak %4d at (%+4d,%4d)    ",value,pkx,pky);
				printf("	(Offset=%3d,Peak=%5d) *\n",off,cf);
#endif
				if ( 0 > LSEGLOCAL.setnewlseg(xctr,pky,pkx, \
							value,off,cf,0,0) ){
					printf("Error: Lseg-local over flow.\n");
					return(2);
				}
			}
#ifdef _Mon_AllPeaks
			else{	printf("Peak %4d at (%+4d,%4d)    ",value,pkx,pky);
				printf("	(Offset=%3d,Peak=%5d)\n",off,cf);
			}
#endif
		}
	}
	time_get(&time_linecorr,0);
	LSEGLOCAL.sortbyoffset();
	LSEGLOCAL.setfirst();
	time_start();  time_areacorr = 0;
	while (1){
		int	x,y,a,htp,off,cf,wu,wl;
		int	af;
		int	id;
		int	acmp,bcmp;
		int	imgoff,imgcf,wucmp,wlcmp;
		if ( -1 == LSEGLOCAL.getnext(&id,&x,&y,&a,&htp,&off,&cf,&wu,&wl) )  break;
		get_vrtwidth(image,xoff,a,y,HrzHist_Bias,&wu,&wl);
						/* imgsamp2 ready */
		af = check_vrthist(wu,wl);
		if ( af ){	/* broad */
			linefit(image,xoff,a,y,wu,wl,&acmp,&bcmp);
			get_vrtwidth(image,xoff,acmp,bcmp,HrzHist_Bias,&wucmp,&wlcmp);
						/* imgsamp2 ready */
		}
		else{	wucmp = wu;  wlcmp = wl;
		}
		mon_saveimage(wucmp,wlcmp);
		af = check_vrthist(wucmp,wlcmp);

		if ( af ){	/* broad */
			int	sum;
			sum = get_imghrzcorr(imgsamp2,ImgSample_Width - wu, \
				ImgSample_Width + wl,imgcorr,imgsamp2->width);
			anl_corr(imgcorr,imgsamp2->width,&imgoff,&imgcf);
			sum = (100 * sum) / (wu + wl + 1) / imgsamp2->width;
			imgcf /= (wu + wl + 1);
//			printf("ImgCorr : (Offset = %4d, Corr = %4d)\n",imgoff,imgcf);
	/* Save Broad-Line which is black enough
	        or         whose auto-correration factor is large. */
			if ( check_BLcorrsum(imgcf,sum) )  af = 0;
		}

#ifndef _Mon_AllPeaks
		if ( infosw ){
			char	*mark = "D";
			if ( cf < 0 ){  cf = -cf;  mark = "C";  }
			printf("%4d  %4d  ",htp,cf);
			if ( (! check_htcorr(htp,cf)) || af ){
				printf("N  Peak %4d at (%4d,%4d,%+4d)    ",htp,x,y,a);
				printf("	(Offset=%3d,Peak=%5d) \n",off,cf);
			}
			else{	printf("%s  Peak %4d at (%4d,%4d,%+4d)    ",mark,htp,x,y,a);
				printf("	(Offset=%3d,Peak=%5d) *\n",off,cf);
			}
		}
#endif
		switch ( mode ){
		    case SW_MODE_ALL:
		    case SW_MODE_TABLE:
			if ( (! check_htcorr(htp,abs(cf))) )  continue;
			break;
		    case 1:
			if ( (! check_htcorr(htp,abs(cf))) || af )  continue;
			break;
		}
		if ( LSEGLST != 0 ){
			if ( 0 > LSEGLST->setnewlseg(x,y,a,htp,off,abs(cf),wu,wl) ){
				printf("Error: Lseg-list over flow.\n");
				return(2);
			}
		}
	}
	time_get(&time_areacorr,0);
	return(0);
}




/*------------------------------------------------------
	Page operation class
------------------------------------------------------*/

class PageOp : public GRPLST {
    private:
	int		merge_groups(LSEGS *LSEGPOOL,int dstgrp,int srcgrp);
	int		clearattr(LSEGS *LSEGPOOL,int attr);
	int		check_lsegconnect(lsegpac *seg1,lsegpac *seg2,int order);
	int		check_lsegconnect2(lsegpac *seg1,lsegpac *seg2,int order);
	int		check_lsegconnect3(lsegpac *seg1,lsegpac *seg2);
	int		grpsortline(int grp,LSEGS *LSEGPOOL);
	int		grpsplit(int grp,LSEGS *LSEGPOOL);
	int		make_conngrp(int grp,LSEGS *LSEGPOOL);
	int		remove_whisker(int grp,LSEGS *LSEGPOOL);
	int		check_match(int v1,int v2,int ori);
	int		remove_paraseg();
	int		combine_lsegs(int grp,LSEGS *LSEGPOOL);
	int		cut_junction();
	int		grpsplit2(int grp,LSEGS *LSEGPOOL);

	int		grpsize;
	int		*grpmatrix;
	int		*grpmatrixrev;
	int		*grpmatrix0;
	int		*grpmatrixrev0;
	int		*grpattr;
	int		grpcreate(int size);
	void		grppushmatrix();
	void		grppopmatrix();
	void		adddedge(int v1,int v2,int code);
	void		removededge(int v1,int v2);
	int		getdedge(int v1,int v2);
	int		getdedgerev(int v1,int v2);
	void		removevertex(int vtx);
	int		getconns(int vtx,int ori);

    public:
	int		mark_group(LSEGS *LSEGPOOL);
	int		delete_alone(LSEGS *LSEGPOOL);
	int		create_groups(LSEGS *LSEGPOOL,int *aslist, \
						int ascounts,int level);
	int		grpsortline_all(LSEGS *LSEGPOOL);
	int		grpsplit_all(LSEGS *LSEGPOOL);
	int		grpsplit2_all(LSEGS *LSEGPOOL);
	int		combine_lsegs_all(LSEGS *LSEGPOOL);

	int		savegroups(char *fname,int width,int height,LSEGS *LSEGPOOL);

	int		print_groups();
	int		print_matrix(int *matrix,int size);
	int		print_matrix(int *matrix,int *matrixrev,int size);

			PageOp();
			~PageOp();
};


PageOp :: PageOp(){
	grpsize = 0;
	grpmatrix = 0;
	grpattr = 0;
}


PageOp :: ~PageOp(){
	if ( grpmatrix != 0 )  delete []grpmatrix;
	if ( grpattr != 0 )  delete []grpattr;
}


int PageOp :: grpcreate(int size){
	int	i;
	if ( grpmatrix != 0 )  delete []grpmatrix;
	if ( grpattr != 0 )  delete []grpattr;
	grpsize = 0;
	if ( 0 == (grpmatrix = new int[4 * size * size]) )  return(-1);
	grpsize = size;
	grpmatrixrev = grpmatrix + (size * size);
	grpmatrix0 = grpmatrix + (2 * size * size);
	grpmatrixrev0 = grpmatrix0 + (size * size);
	if ( 0 == (grpattr = new int[size]) )  return(-1);
	for ( i=0 ; i < (size * size) ; i++ )  grpmatrix[i] = grpmatrixrev[i] = 0;
	for ( i=0 ; i < size ; i++ )  grpattr[i] = 0;
	return(0);
}


void PageOp :: grppushmatrix(){
	memcpy((void *)grpmatrix0,(void *)grpmatrix,sizeof(int) * grpsize * grpsize);
	memcpy((void *)grpmatrixrev0,(void *)grpmatrixrev,sizeof(int) * grpsize * grpsize);
}


void PageOp :: grppopmatrix(){
	memcpy((void *)grpmatrix,(void *)grpmatrix0,sizeof(int) * grpsize * grpsize);
	memcpy((void *)grpmatrixrev,(void *)grpmatrixrev0,sizeof(int) * grpsize * grpsize);
}


void PageOp :: adddedge(int v1,int v2,int code){
	grpmatrix[v2 + (v1 * grpsize)] = code;
	grpmatrixrev[v1 + (v2 * grpsize)] = code;
}


void PageOp :: removededge(int v1,int v2){
	grpmatrix[v2 + (v1 * grpsize)] = 0;
	grpmatrixrev[v1 + (v2 * grpsize)] = 0;
}


int PageOp :: getdedge(int v1,int v2){
	return ( grpmatrix[v2 + (v1 * grpsize)] );
}


int PageOp :: getdedgerev(int v1,int v2){
	return ( grpmatrixrev[v2 + (v1 * grpsize)] );
}


void PageOp :: removevertex(int vtx){
	int	i;
	for ( i=0 ; i<grpsize ; i++ ){
		grpmatrix   [i + (vtx * grpsize)] = 0;
		grpmatrixrev[i + (vtx * grpsize)] = 0;
		grpmatrix   [vtx + (i * grpsize)] = 0;
		grpmatrixrev[vtx + (i * grpsize)] = 0;
	}
}


int PageOp :: getconns(int vtx,int ori){
	int	i;
	int	cf,cb;
	cf = cb = 0;
	for ( i=0 ; i<grpsize ; i++ ){
		if ( grpmatrix   [i + (vtx * grpsize)] != 0 )  ++cf;
		if ( grpmatrixrev[i + (vtx * grpsize)] != 0 )  ++cb;
	}
	switch(ori){
	    case 0:	return(cf + cb);
	    case 1:	return(cf);
	    case -1:	return(cb);
	}
	return(0);
}




int PageOp :: merge_groups(LSEGS *LSEGPOOL,int dstgrp,int srcgrp){
	int	i,idcounts;
	int	*idlst;
	idlst = getidlist(srcgrp);
	idcounts = getidcounts(srcgrp);
	for ( i=0 ; i<idcounts ; i++ ){
		LSEGPOOL->setattr(idlst[i],dstgrp);
	}
	if ( mergegrps(dstgrp,srcgrp) )  return(-1);
	return(0);
}


int PageOp :: clearattr(LSEGS *LSEGPOOL,int attr){
	int	i;
	for ( i=0 ; i < LSEGPOOL->counts ; i++ )  LSEGPOOL->setattr(i,attr);
	return(0);
}


int PageOp :: check_lsegconnect(lsegpac *seg1,lsegpac *seg2,int order){
	int	xdiff,yd1,yd2,y1,y2;
	int	h1,h2;
	int	an1,an2;
	xdiff = seg2->x - seg1->x;
	if ( abs(xdiff) != (BWID /2) )  return(0);
	an1 = atan_deg(seg1->a);
	an2 = atan_deg(seg2->a);
//	if ( abs(seg2->a - seg1->a) > ((3 * AngleReso) /32) )  return(0);
							/* 5.4deg */
	if ( abs(an2 - an1) > (10 * DegMul) )  return(0);
	if ( order && ((order * xdiff) < 0) )  return(0);
	if ( xdiff > 0 ){	/*  seg1 : seg2  */
		y1 = -((BWID /2) * seg1->a) / AngleReso + seg1->y;
		y2 =  ((BWID /2) * seg2->a) / AngleReso + seg2->y;
		yd1 = abs(y2 - seg1->y);
		yd2 = abs(y1 - seg2->y);
	}
	else{			/*  seg2 : seg1  */
		y1 =  ((BWID /2) * seg1->a) / AngleReso + seg1->y;
		y2 = -((BWID /2) * seg2->a) / AngleReso + seg2->y;
		yd1 = abs(y1 - seg2->y);
		yd2 = abs(y2 - seg1->y);
	}
	if ( yd1 > yd2 )  yd1 = yd2;	/* min */
#if 0
	if ( yd1 > 4 )  return(0);
#else
	h1 = seg1->wu + seg1->wl +1;
	h2 = seg2->wu + seg2->wl +1;
	if ( h1 < h2 )  h1 = h2;
	if ( h1 < 8 )  h1 = 8;
	h1 /= 2;
	if ( yd1 > h1 )  return(0);
#endif

	if ( xdiff > 0 )  return(1);
	return(-1);
}


int PageOp :: check_lsegconnect2(lsegpac *seg1,lsegpac *seg2,int order){
	int	xdiff,y1,y2;
	int	h1,h2;
	int	an1,an2;
	xdiff = seg2->x - seg1->x;
	if ( abs(xdiff) != BWID )  return(0);
	an1 = atan_deg(seg1->a);
	an2 = atan_deg(seg2->a);
//	if ( abs(seg2->a - seg1->a) > ((3 * AngleReso) /32) )  return(0);
							/* 5.4deg */
	if ( abs(an2 - an1) > (10 * DegMul) )  return(0);
	if ( order && ((order * xdiff) < 0) )  return(0);
	if ( xdiff > 0 ){	/*  seg1 : seg2  */
		y1 = -((BWID /2) * seg1->a) / AngleReso + seg1->y;
		y2 =  ((BWID /2) * seg2->a) / AngleReso + seg2->y;
	}
	else{			/*  seg2 : seg1  */
		y1 =  ((BWID /2) * seg1->a) / AngleReso + seg1->y;
		y2 = -((BWID /2) * seg2->a) / AngleReso + seg2->y;
	}
#if 0
	if ( abs(y1 - y2) > 4 )  return(0);
#else
	h1 = seg1->wu + seg1->wl +1;
	h2 = seg2->wu + seg2->wl +1;
	if ( h1 < h2 )  h1 = h2;
	if ( h1 < 8 )  h1 = 8;
	h1 /= 2;
	if ( abs(y1 - y2) > h1 )  return(0);
#endif

	if ( xdiff > 0 )  return(1);
	return(-1);
}


int PageOp :: check_lsegconnect3(lsegpac *seg1,lsegpac *seg2){
	int	xdiff;
	int	an1,an2;
	double	co;
	int	xc,yc,rx1,ry1,rx2,ry2;
	xdiff = seg2->x - seg1->x;
	if ( abs(xdiff) > BWID )  return(0);
	an1 = atan_deg(seg1->a);
	an2 = atan_deg(seg2->a);
	if ( abs(an2 - an1) > (10 * DegMul) )  return(0);
	co = cos(atan((double)seg1->a / AngleReso)); 
#if 0
	int	y1,y2;
	y1 = seg1->y;
	y2 = seg2->y;
	if ( (y2 - y1) >= 0 ){
		if ( (an1 * xdiff) > 0 )  return(0);
		if ( (y2 - (int)((double)seg2->wu * co)) < y1 )  return(-1);
		if ( (y1 + (int)((double)seg1->wl * co)) > y2 )  return(-1);
	}
	else{
		if ( (an1 * xdiff) < 0 )  return(0);
		if ( (y1 - (int)((double)seg1->wu * co)) < y2 )  return(-1);
		if ( (y2 + (int)((double)seg2->wl * co)) > y1 )  return(-1);
	}
#else
	rx1 = seg1->x;	ry1 = seg1->y;
	rx2 = seg2->x;	ry2 = seg2->y;
	xc = (rx1 + rx2) /2;
	yc = (ry1 + ry2) /2;
	rx1 -= xc;	ry1 -= yc;	ry1 = -ry1;
	rx2 -= xc;	ry2 -= yc;	ry2 = -ry2;
	rotate(-seg1->a,&rx1,&ry1);
	rotate(-seg1->a,&rx2,&ry2);
	xdiff = rx2 - rx1;
	if ( abs(xdiff) > (BWID /2) )  return(0);
	ry2 = 2 * abs(ry1);
	if ( ry1 >= 0 ){
		if ( ry2 <= (int)((double)seg1->wl * co) )  return(-1);
		if ( ry2 <= (int)((double)seg2->wu * co) )  return(-1);
	}
	else{
		if ( ry2 <= (int)((double)seg1->wu * co) )  return(-1);
		if ( ry2 <= (int)((double)seg2->wl * co) )  return(-1);
	}
#endif
	return(0);
}


int PageOp :: grpsortline(int grp,LSEGS *LSEGPOOL){
	int	*idtmp,id0;
	int	i,j;
	int	idcounts;
	lsegpac	*seg1,*seg2;
	if ( (idcounts = getidcounts(grp)) == 0 )  return(0);
	if ( 0 == (idtmp = new int[idcounts]) )  return(-1);
	memcpy((void *)idtmp,(void *)getidlist(grp),sizeof(int) * idcounts);
	for ( j=0 ; j<idcounts ; j++ ){
		for ( i=j+1 ; i<idcounts ; i++ ){
			seg1 = LSEGPOOL->getlseg(idtmp[i]);
			seg2 = LSEGPOOL->getlseg(idtmp[j]);
			if ( seg1->x < seg2->x ){
				id0 = idtmp[i];
				idtmp[i] = idtmp[j];
				idtmp[j] = id0;
			}
			else{	if ( (seg1->x == seg2->x) && (seg1->y < seg2->y) ){
					id0 = idtmp[i];
					idtmp[i] = idtmp[j];
					idtmp[j] = id0;
				}
			}
		}
	}
	memcpy((void *)getidlist(grp),(void *)idtmp,sizeof(int) * idcounts);
	delete []idtmp;
	return(0);
}


int PageOp :: mark_group(LSEGS *LSEGPOOL){
	int		s1,s2;
	lsegpac		*seg1,*seg2;
	for ( s2=0 ; s2 < LSEGPOOL->counts ; s2++ )  LSEGPOOL->setattr(s2,-1);
	for ( s2=0 ; s2 < LSEGPOOL->counts ; s2++ ){
		if ( ! LSEGPOOL->isactive(s2) )  continue;
		seg2 = LSEGPOOL->getlseg(s2);
		for ( s1=s2+1 ; s1 < LSEGPOOL->counts ; s1++ ){
			if ( ! LSEGPOOL->isactive(s1) )  continue;
			seg1 = LSEGPOOL->getlseg(s1);
			if ( ! check_lsegconnect(seg1,seg2,0) )  continue;
			LSEGPOOL->setattr(s1,1);
			LSEGPOOL->setattr(s2,1);
		}
	}
	return(0);
}


int PageOp :: delete_alone(LSEGS *LSEGPOOL){
	int		sn;
	lsegpac		*lseg;
	int		grp;
	for ( sn=0 ; sn < LSEGPOOL->counts ; sn++ ){
		if ( ! LSEGPOOL->isactive(sn) )  continue;
#if 0
		if ( LSEGPOOL->getattr(sn) != -1 )  continue;
#else
		grp = LSEGPOOL->getattr(sn);
		if ( getidcounts(grp) > 1 )  continue;
#endif
		lseg = LSEGPOOL->getlseg(sn);
//		if ( ! check_vrthist(lseg->wu,lseg->wl) )  continue;
		if ( ! check_vrthist_del(lseg->wu,lseg->wl) )  continue;
		LSEGPOOL->deletelseg(sn);
	}
	return(0);
}


int PageOp :: create_groups(LSEGS *LSEGPOOL,int *aslist,int ascounts,int level){
	int		i,j,s1,s2;
	lsegpac		*seg1,*seg2;
	int		grp1,grp2;
	int		asegs,*aseglist;
	int		retcode;
	clearattr(LSEGPOOL,-1);
	if ( aslist == 0 ){
		if ( 0 == (aseglist = new int[LSEGPOOL->counts]) )  return(-1);
		for ( i=asegs=0 ; i < LSEGPOOL->counts ; i++ ){
			if ( ! LSEGPOOL->isactive(i) )  continue;
			aseglist[asegs++] = i;
		}
	}
	else{	asegs = ascounts;
		aseglist = aslist;
	}
	retcode = 0;
	for ( j=0 ; (j < asegs) && (retcode == 0) ; j++ ){
		s2 = aseglist[j];
		grp2 = LSEGPOOL->getattr(s2);
		if ( -1 == grp2 ){
			if ( (grp2 = newgrp(s2)) < 0 ){ retcode = -1;  break; }
			LSEGPOOL->setattr(s2,grp2);
		}
		seg2 = LSEGPOOL->getlseg(s2);
		for ( i=0 ; i<asegs ; i++ ){
			s1 = aseglist[i];
			if ( ! LSEGPOOL->isactive(s1) )  continue;
			grp1 = LSEGPOOL->getattr(s1);
			seg1 = LSEGPOOL->getlseg(s1);
			if ( level == 0 ){
				if ( ! check_lsegconnect(seg1,seg2,0) )  continue;
			}
			else{	if (    (! check_lsegconnect(seg1,seg2,0)) \
				     && (! check_lsegconnect2(seg1,seg2,0)) \
				     && (! check_lsegconnect3(seg1,seg2)) )  continue;
			}
			if ( -1 != grp1 ){
				if ( grp1 == grp2 )  continue;
				if ( merge_groups(LSEGPOOL,grp2,grp1) ){
					retcode = -2;  break;
				}
			}
			else{	if ( addgrp(grp2,s1) ){  retcode = 3;  break; }
				LSEGPOOL->setattr(s1,grp2);
			}
		}
	}
	if ( aslist == 0 )  delete []aseglist;
	return(retcode);
}


int PageOp :: grpsortline_all(LSEGS *LSEGPOOL){
	int	grp;
	for ( grp=0 ; grp<groups ; grp++ ){
		if ( grpsortline(grp,LSEGPOOL) )  return(-1);
	}
	return(0);
}


int PageOp :: grpsplit(int grp,LSEGS *LSEGPOOL){
	int	i1,i2;
	lsegpac	*seg1,*seg2;
	int	idcounts;
	int	*idlist;
	int	*clist;
	int	cf;
	idcounts = getidcounts(grp);
	if ( idcounts <= 2 )  return(0);
	idlist = getidlist(grp);
	if ( 0 == (clist = new int[idcounts * idcounts]) )  return(-1);
	for ( i1=0; i1<idcounts ; i1++ ){
		for ( i2=0; i2<idcounts ; i2++ )  clist[i2 + (i1 * idcounts)] = 0;
	}
	for ( i1=0; i1<idcounts ; i1++ ){
		seg1 = LSEGPOOL->getlseg(idlist[i1]);
		cf = 0;
		for ( i2=0; i2<idcounts ; i2++ ){	/* Left 1 */
			seg2 = LSEGPOOL->getlseg(idlist[i2]);
			if ( ! check_lsegconnect(seg1,seg2,-1) )  continue;
			cf = -1;
			clist[i2 + (i1 * idcounts)] = 3;
			clist[i1 + (i2 * idcounts)] = 2;
		}
		if ( cf == 0 ){
			for ( i2=0; i2<idcounts ; i2++ ){	/* Left 2 */
				seg2 = LSEGPOOL->getlseg(idlist[i2]);
				if ( ! check_lsegconnect2(seg1,seg2,-1) )  continue;
				cf = -1;
				clist[i2 + (i1 * idcounts)] = 5;
				clist[i1 + (i2 * idcounts)] = 4;
			}
		}
		cf = 0;
		for ( i2=0; i2<idcounts ; i2++ ){	/* Right 1 */
			seg2 = LSEGPOOL->getlseg(idlist[i2]);
			if ( ! check_lsegconnect(seg1,seg2,1) )  continue;
			cf = -1;
			clist[i2 + (i1 * idcounts)] = 2;
			clist[i1 + (i2 * idcounts)] = 3;
		}
		if ( cf == 0 ){
			for ( i2=0; i2<idcounts ; i2++ ){	/* Right 2 */
				seg2 = LSEGPOOL->getlseg(idlist[i2]);
				if ( ! check_lsegconnect2(seg1,seg2,1) )  continue;
				cf = -1;
				clist[i2 + (i1 * idcounts)] = 4;
				clist[i1 + (i2 * idcounts)] = 5;
			}
		}
	}
	if ( idcounts < 60 ){
		printf("Group = %d :\n",grp);
		print_matrix(clist,idcounts);
	}
	delete []clist;
	return(0);
}


int PageOp :: grpsplit_all(LSEGS *LSEGPOOL){
	int	grp;
	for ( grp=0 ; grp<groups ; grp++ ){
		if ( grpsplit(grp,LSEGPOOL) )  return(-1);
	}
	return(0);
}


int PageOp :: combine_lsegs_all(LSEGS *LSEGPOOL){
	int	grp;
	for ( grp=0 ; grp<groups ; grp++ ){
		if ( combine_lsegs(grp,LSEGPOOL) )  return(-1);
	}
	return(0);
}


int PageOp :: make_conngrp(int grp,LSEGS *LSEGPOOL){
	GRPLST	GRP;
	int	i,i1,i2,s1,s2,grp1,grp2;
	int	idcounts;
	int	*idlist;
	int	actgrps;
	lsegpac	*seg1,*seg2;
	int	conorder;
	idcounts = getidcounts(grp);
//	if ( idcounts < 2 )  return(0);
//	if ( idcounts <= 2 )  return(0);
	idlist = getidlist(grp);
	if ( 0 != GRP.alloc(idcounts) )  return(-1);
	if ( grpcreate(idcounts) )  return(-1);
	for ( i1=0; i1<idcounts ; i1++ )  LSEGPOOL->setattr(idlist[i1],-1);

	for ( i1=0; i1<idcounts ; i1++ ){		/* Raw */
		s1 = idlist[i1];
		grp1 = LSEGPOOL->getattr(s1);
		if ( -1 == grp1 ){
			if ( 0 > (grp1 = GRP.newgrp(s1)) )  return(-1);
			LSEGPOOL->setattr(s1,grp1);
		}
		seg1 = LSEGPOOL->getlseg(s1);
		for ( i2=0; i2<idcounts ; i2++ ){	/* Column */
			s2 = idlist[i2];
			grp2 = LSEGPOOL->getattr(s2);
			seg2 = LSEGPOOL->getlseg(s2);
			if ( ! (conorder = check_lsegconnect(seg1,seg2,0)) )  continue;
			if ( conorder > 0 )  adddedge(i1,i2,1);
			else                 adddedge(i2,i1,1);
			if ( -1 != grp2 ){
				if ( grp1 == grp2 )  continue;
				int	*idlst,idcounts;
				idlst = GRP.getidlist(grp2);
				idcounts = GRP.getidcounts(grp2);
				for ( i=0 ; i<idcounts ; i++ ){
					LSEGPOOL->setattr(idlst[i],grp1);
				}
				if ( GRP.mergegrps(grp1,grp2) )  return(-2);
			}
			else{	if ( GRP.addgrp(grp1,s2) )  return(-3);
				LSEGPOOL->setattr(s2,grp2);
			}
		}
	}

	for ( i=actgrps=0 ; i<GRP.groups ; i++ ){
		if ( GRP.getidcounts(i) != 0 )  ++actgrps;
	}
	printf("In group %d :  %d active sub-groups\n",grp,actgrps);

	for ( i1=0; i1<idcounts ; i1++ ){		/* Raw */
		s1 = idlist[i1];
		grp1 = LSEGPOOL->getattr(s1);
		seg1 = LSEGPOOL->getlseg(s1);
		for ( i2=0; i2<idcounts ; i2++ ){	/* Column */
			s2 = idlist[i2];
			grp2 = LSEGPOOL->getattr(s2);
			seg2 = LSEGPOOL->getlseg(s2);
			if ( ! (conorder = check_lsegconnect2(seg1,seg2,0)) )  continue;
			if ( grp1 == grp2 )  continue;
			if ( conorder > 0 )  adddedge(i1,i2,2);
			else                 adddedge(i2,i1,2);
		}
	}

	return(0);
}


//#define	_Whisker_Remove

int PageOp :: remove_whisker(int grp,LSEGS *LSEGPOOL){
	int	i,v1,v2;
	int	idcounts;
	int	*idlist;
	int	cleft,cright;
	int	rmvf;
	idcounts = grpsize;	/* = getidcounts(grp) */
	idlist = getidlist(grp);
	rmvf = 0;
	for ( v1=0 ; v1<idcounts ; v1++ ){
		cleft  = getconns(v1,-1);
		cright = getconns(v1,1);
//		printf("Conn : %d - %d\n",cleft,cright);
		if ( cleft  > 1 )  continue;
		if ( cright > 1 )  continue;
		if ( (cleft > 0) && (cright > 0) )  continue;
		if ( (cleft == 0) && (cright == 0) )  continue;
		if ( cleft != 0 ){	/*  -O  */
			for ( v2=0 ; v2<grpsize ; v2++ ){
				if ( 0 != getdedge(v2,v1) )  break;
			}
			for ( i=0 ; i<grpsize ; i++ ){
				if ( 0 == grpmatrix[i + (v2 * grpsize)] )  continue;
				if ( i == v1 )  continue;
				if ( getconns(i,1) > 0 ){
					printf("ID. %d --- whisker\n",v1);
					grpattr[v1] = 1;
					removevertex(v1);
#ifdef	_Whisker_Remove
					LSEGPOOL->deletelseg(idlist[v1]);
					idlist[v1] = -1;
					rmvf = 1;
#endif

					break;
				}
			}
		}
		else{			/*  O-  */
			for ( v2=0 ; v2<grpsize ; v2++ ){
				if ( 0 != getdedge(v1,v2) )  break;
			}
			for ( i=0 ; i<grpsize ; i++ ){
				if ( 0 == grpmatrixrev[i + (v2 * grpsize)] )  continue;
				if ( i == v1 )  continue;
				if ( getconns(i,-1) > 0 ){
					printf("ID. %d --- whisker\n",v1);
					grpattr[v1] = 1;
					removevertex(v1);
#ifdef	_Whisker_Remove
					LSEGPOOL->deletelseg(idlist[v1]);
					idlist[v1] = -1;
					rmvf = 1;
#endif
					break;
				}
			}
		}
	}
	if ( rmvf ){
		v1 = idcounts;  idcounts = 0;
		for ( i=0 ; i<v1 ; i++ ){
			if ( idlist[i] == -1 )  continue;
			idlist[ idcounts++ ] = idlist[i];
		}
		setidcounts(grp,idcounts);
		make_conngrp(grp,LSEGPOOL);
		grppushmatrix();
	}
	return(0);
}


int PageOp :: check_match(int v1,int v2,int ori){
	int	i,j;
	if ( ori > 0 ){		/* Right */
		for ( i=0 ; i<grpsize ; i++ ){
			if ( 0 == grpmatrix[i + (v1 * grpsize)] )  continue;
			for ( j=0 ; j<grpsize ; j++ ){
				if ( 0 == grpmatrix[j + (v2 * grpsize)] )  continue;
				if ( i == j )  return(i);	/* Matched */
			}
		}
	}
	else{			/* Left */
		for ( i=0 ; i<grpsize ; i++ ){
			if ( 0 == grpmatrixrev[i + (v1 * grpsize)] )  continue;
			for ( j=0 ; j<grpsize ; j++ ){
				if ( 0 == grpmatrixrev[j + (v2 * grpsize)] )  continue;
				if ( i == j )  return(i);	/* Matched */
			}
		}
	}
	return(-1);	/* No match */
}


int PageOp :: remove_paraseg(){
	int	i,j,v1;
	int	idcounts;
	int	cleft,cright;
	int	*attnid,mp;
	int	attnsegs;
	idcounts = grpsize;
	if ( 0 == (attnid = new int[idcounts]) )  return(-1);
	for ( v1=0 ; v1<idcounts ; v1++ ){
		cleft  = getconns(v1,-1);
		cright = getconns(v1,1);
		if ( (cleft <= 1) && (cright <= 1) )  continue;
		if ( cleft > 1 ){
			attnsegs = 0;
			for ( i=0 ; i<idcounts ; i++ ){
				if ( 0 != grpmatrixrev[i + (v1 * grpsize)] ){
					attnid[ attnsegs++ ] = i;
				}
			}
			for ( i=0 ; i<attnsegs ; i++ ){
				for ( j=i+1 ; j<attnsegs ; j++ ){
					mp = check_match(attnid[i],attnid[j],-1);
					if ( -1 == mp )  continue;
					printf("ID. %d --- paraseg\n",attnid[j]);
					grpattr[attnid[j]] = 2;
					removevertex(attnid[j]);
				}
			}
		}
		if ( cright > 1 ){
			attnsegs = 0;
			for ( i=0 ; i<idcounts ; i++ ){
				if ( 0 != grpmatrix[i + (v1 * grpsize)] ){
					attnid[ attnsegs++ ] = i;
				}
			}
			for ( i=0 ; i<attnsegs ; i++ ){
				for ( j=i+1 ; j<attnsegs ; j++ ){
					mp = check_match(attnid[i],attnid[j],1);
					if ( -1 == mp )  continue;
					printf("ID. %d --- paraseg\n",attnid[j]);
					grpattr[attnid[j]] = 2;
					removevertex(attnid[j]);
				}
			}
		}
	}
	delete []attnid;
	return(0);
}


int PageOp :: combine_lsegs(int grp,LSEGS *LSEGPOOL){
	int	i,j;
	int	idcounts,ids;
	int	*idlist;
	int	x0,x;
	int	ysum,a;
	idcounts = getidcounts(grp);
	idlist = getidlist(grp);
	ids = 0;
	for ( i=0 ; i<idcounts ; i++ ){
		x0 = LSEGPOOL->getlseg(idlist[i])->x;
		ysum = LSEGPOOL->getlseg(idlist[i])->y;
		for ( j=i+1 ; j<idcounts ; j++ ){
			x = LSEGPOOL->getlseg(idlist[j])->x;
			if ( x0 != x )  break;
			ysum += LSEGPOOL->getlseg(idlist[j])->y;
			LSEGPOOL->deletelseg(idlist[j]);
		}
		ysum /= (j - i);
		LSEGPOOL->getlseg(idlist[i])->y = ysum;
		idlist[ ids++ ] = idlist[i];
		i = j - 1;
	}
	setidcounts(grp,ids);
	for ( i=1 ; i < (ids - 1) ; i++ ){
		x  = LSEGPOOL->getlseg(idlist[i+1])->x;
		x -= LSEGPOOL->getlseg(idlist[i-1])->x;
		a  = LSEGPOOL->getlseg(idlist[i+1])->y;
		a -= LSEGPOOL->getlseg(idlist[i-1])->y;
		a = -a * AngleReso / x;
		LSEGPOOL->getlseg(idlist[i])->a = a;
	}
	return(0);
}


int PageOp :: cut_junction(){
	int	i,v1;
	for ( v1=0 ; v1<grpsize ; v1++ ){
		if ( getconns(v1,-1) > 1 ){	/* Left */
			for ( i=0 ; i<grpsize ; i++ ){
				if ( grpmatrixrev[i + (v1 * grpsize)] != 0 ){
					grpmatrixrev[i + (v1 * grpsize)] |= 0x80;
					grpmatrix[v1 + (i * grpsize)] |= 0x80;
					printf("Cut junction %d - %d\n",v1,i);
				}
			}
		}
		if ( getconns(v1,1) > 1 ){	/* Right */
			for ( i=0 ; i<grpsize ; i++ ){
				if ( grpmatrix[i + (v1 * grpsize)] != 0 ){
					grpmatrix[i + (v1 * grpsize)] |= 0x80;
					grpmatrixrev[v1 + (i * grpsize)] |= 0x80;
					printf("Cut junction %d - %d\n",v1,i);
				}
			}
		}
	}
	return(0);
}


int PageOp :: grpsplit2(int grp,LSEGS *LSEGPOOL){
	int	idcounts;
	idcounts = getidcounts(grp);
	if ( make_conngrp(grp,LSEGPOOL) )  return(-1);
	grppushmatrix();

#if 0
	if ( idcounts < 60 ){
		printf("Group = %d :\n",grp);
		print_matrix(grpmatrix,grpmatrixrev,idcounts);
	}
#endif

	remove_paraseg();
	remove_whisker(grp,LSEGPOOL);

#if 0
	if ( idcounts < 60 ){
		printf("Group = %d :\n",grp);
		print_matrix(grpmatrix,grpmatrixrev,idcounts);
	}
#endif

	cut_junction();

#if 1
	if ( idcounts < 60 ){
		printf("Group = %d :\n",grp);
		print_matrix(grpmatrix,grpmatrixrev,idcounts);
	}
#endif

	return(0);
}


int PageOp :: grpsplit2_all(LSEGS *LSEGPOOL){
	int	grp;
	for ( grp=0 ; grp<groups ; grp++ ){
		if ( grpsplit2(grp,LSEGPOOL) )  return(-1);
	}
	return(0);
}


int PageOp :: savegroups(char *fname,int width,int height,LSEGS *LSEGPOOL){
	int	i;
	d8file	d8h;
	int	grp;
	int	idcounts;
	int	*idlist;
	lsegpac	*lsp;
	D8SAVE	D8SV;
	struct d8_lseg{
		short	x,y;
		short	a;
		short	htpeak;
		short	corroff,corrpeak;
		short	wu,wl;
	};
	d8_lseg	d8data;
	d8_lseg	d8term = { 0,0,0,0, 0,0,0,0 };

	for ( i=0 ; i<8 ; i++ )  d8h.d8_llimit[i] = d8h.d8_hlimit[i] = 0;
	d8h.d8_hlimit[0] = width -1;
	d8h.d8_hlimit[1] = height -1;
	d8h.d8_llimit[2] = -AngleReso;
	d8h.d8_hlimit[2] = AngleReso;
	d8h.d8_hlimit[3] = 32767;
	d8h.d8_llimit[4] = -1;
	d8h.d8_hlimit[4] = BWID;
	d8h.d8_hlimit[5] = 32767;
	d8h.d8_hlimit[6] = ImgSample_Width +1;
	d8h.d8_hlimit[7] = ImgSample_Width +1;
	if ( D8SV.creat(fname,&d8h) )  return(-1);
	for ( grp=0 ; grp < groups ; grp++ ){
		idcounts = getidcounts(grp);
		if ( idcounts == 0 )  continue;
		idlist = getidlist(grp);
		for ( i=0 ; i < idcounts ; i++ ){
			if ( 0 == (lsp = LSEGPOOL->getlseg(idlist[i])) )  continue;
			d8data.x = (short)lsp->x;
			d8data.y = (short)lsp->y;
			d8data.a = (short)lsp->a;
			d8data.htpeak = (short)lsp->htpeak;
			d8data.corroff = (short)lsp->corroff;
			d8data.corrpeak = (short)lsp->corrpeak;
			d8data.wu = (short)lsp->wu;
			d8data.wl = (short)lsp->wl;
			if ( D8SV.putdata((d8pac *)&d8data) ){
				D8SV.close();
				unlink(fname);
				return(-2);
			}
		}
		if ( D8SV.putdata((d8pac *)&d8term) ){
			D8SV.close();
			unlink(fname);
			return(-2);
		}
	}
	if ( D8SV.close() )  return(-1);
	return(0);
}


int PageOp :: print_groups(){
	int	i,grp;
	int	idcounts;
	int	*idlist;
	for ( grp=0 ; grp < groups ; grp++ ){
		idcounts = getidcounts(grp);
		if ( idcounts == 0 )  continue;
		idlist = getidlist(grp);
		printf("Group %3d : ",grp);
		for ( i=0 ; i < idcounts ; i++ ){
			printf("%d  ",idlist[i]);
		}
		printf("\n");
	}
	return(0);
}


int PageOp :: print_matrix(int *matrix,int size){
	int	i1,i2;
	for ( i1=0 ; i1<size ; i1++ ){
		printf("    ");
		for ( i2=0 ; i2<size ; i2++ ){
			switch( matrix[i2 + (i1 * size)] ){
			    case 0:	printf("- ");  break;
			    case 1:	printf("1 ");  break;
			    case 2:	printf("2 ");  break;
			    case 3:	printf("3 ");  break;
			    case 4:	printf("4 ");  break;
			    case 5:	printf("5 ");  break;
			}
		}
		printf("\n");
	}
	return(0);
}


int PageOp :: print_matrix(int *matrix,int *matrixrev,int size){
	int	i1,i2;
	int	cf,cb;
	int	mask;
	for ( i1=0 ; i1<size ; i1++ ){
		printf("    %3d: ",i1);
		for ( i2=0 ; i2<size ; i2++ ){
			cf = matrix[i2 + (i1 * size)];
			cb = matrixrev[i2 + (i1 * size)];
			if ( (cb == 0) && (cf == 0) ){
				printf("- ");  continue;
			}
			mask  = cf & 0x80;	cf &= ~0x80;
			mask |= cb & 0x80;	cb &= ~0x80;
			cf *= 2;
			cb *= 2;
			if ( cb != 0 )  cf = cb + 1;
			switch(cf){
			    case 1:	printf("1");  break;
			    case 2:	printf("2");  break;
			    case 3:	printf("3");  break;
			    case 4:	printf("4");  break;
			    case 5:	printf("5");  break;
			}
			if ( mask == 0 )  printf(" ");  else printf("'");
		}
		printf("\n");
	}
	return(0);
}




/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
};
    	
    	
void MSG :: usage(void){		// Display Usage
	puts("getlseg  v2.1, Copyright (C) 1994-1998 Hideaki Goto ");
	puts("Usage : getlseg [-options] in_file|stdin outfile");
	puts("          -all     : get all line segments ");
	puts("       *  -rule    : rule extraction ");
	puts("          -table   : rule extraction (table processing) ");
	puts("          -r       : read image rotated by 90degs clockwise");
	puts("          -h       : get horizontal line segments only ");
	puts("          -info    : show operation information ");
#ifndef	HAS_NO_TIMES
	puts("          -time    : show operation time ");
#endif
//	puts("          -quiet   : suppress messages ");
}


void MSG :: switcherr(void){
	puts("Switch error.");
}


void MSG :: readerr(char *name){
	printf("Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	printf("Can't write %s.\n",name);
}


void MSG :: nonsupported(char *name){
	printf("Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	puts("Not enough memory.");
}




/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/


class ARGS {
    private:
	char		*fpath[16];
	int		fpcount;
    	int		sw_verbose;
	int		sw_info;
	int		sw_timeinfo;
	int		sw_mode;
	int		sw_direct;
	int		sw_thin;
	int		sw_hw;
	int		sw_rotate;
    public:
	int		scanargv(int ac,char **av);
	int		SWverbose(void) {  return sw_verbose; }
	int		SWinfo(void) {  return sw_info; }
	int		SWtimeinfo(void) {  return sw_timeinfo; }
	int		SWmode(void) {  return sw_mode; }
	int		SWdirect(void) {  return sw_direct; }
	int		SWthin(void) {  return sw_thin; }
	int		SWhw(void) {  return sw_hw; }
	int		SWrotate(void) {  return sw_rotate; }
	char		*infile(void);
	char		*outfile(void);
			ARGS(void);
};


ARGS :: ARGS(void){
	sw_verbose = -1;
	sw_info = 0;
	sw_timeinfo = 0;
	sw_mode = SW_MODE_RULE;
	sw_direct = 0;
	sw_thin = 0;
	sw_hw = -1;
	sw_rotate = 0;
}


int ARGS :: scanargv(int ac,char **av){
	int	i1,err,skip;
	char	*sw;
	fpcount = err = 0;
	skip = 0;
	--ac;  if( ac >= 16 )  ac = 16;
	for( i1 = 0 ; i1 < ac ; ++i1 ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i1+1] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"h") ){
				sw_hw = 2;
				continue;
			}
			if( 0 == strcmp(sw,"thin") ){
				sw_thin = -1;
				continue;
			}
			if( 0 == strcmp(sw,"all") ){
				sw_mode = SW_MODE_ALL;
				continue;
			}
			if( 0 == strcmp(sw,"rule") ){
				sw_mode = SW_MODE_RULE;
				continue;
			}
			if( 0 == strcmp(sw,"table") ){
				sw_mode = SW_MODE_TABLE;
				continue;
			}
			if( 0 == strcmp(sw,"info") ){
				sw_info = -1;
				continue;
			}
			if( 0 == strcmp(sw,"time") ){
				sw_timeinfo = -1;
				continue;
			}
			if( 0 == strcmp(sw,"direct") ){
				sw_direct = -1;
				continue;
			}
			if( 0 == strcmp(sw,"r") ){
				sw_rotate = -1;
				continue;
			}
			if( 0 == strcmp(sw,"quiet") ){
				sw_verbose = 0;
				continue;
			}
			if( 0 == strcmp(sw,"verbose") ){
				sw_verbose = -1;
				continue;
			}
			err = -1;  break;
		}
		else{	fpath[fpcount++] = av[i1+1];
		}
	}
	return(err);
}


char * ARGS :: infile(void){
	if ( fpcount < 1 )  return( "" );
	return( fpath[0] );
}


char * ARGS :: outfile(void){
	if ( fpcount < 2 )  return( "tmp.getlseg" );
	return( fpath[1] );
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/

int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	int		retcode;
	int		xoff;
	char		*outfile;
	BandOp		BandOp;
	PageOp		PageOp;
	int		width,height;
	SIPImage	*image;
	SIPImage	*thinimage;
	LSEGPOOL	LSEGPOOL;
	double		time_thinning;
	double		time_rlencode,time_hough,time_peakdet,time_lsegcomp;
	double		time_linecorr,time_areacorr,time_rulegrouping;

	if ( ac < 2 ){
		MSG.usage();
		return(0);
	}
	if ( ARGS.scanargv(ac,av) ){
		MSG.switcherr();
		return(1);
	}
	outfile = ARGS.outfile();


//---------------------	Read image file

	retcode = 0;
	if ( 0 == (image = loadimage(ARGS.infile(),&width,&height, \
			ARGS.SWrotate(),ARGS.SWinfo())) ){
		printf("Can\'t read %s. \n",ARGS.infile());
		return(1);
	}

	if ( 0 != BandOp.init(BWID,height) ){
		printf("Memory not enough. \n");
		retcode = 1;
		goto main_exit;
	}


//---------------------	Create thinned image

	thinimage = 0;
	time_thinning = 0;
	retcode = 0;
	if ( ARGS.SWthin() ){
		int		i;
		SIPRectangle	rect;
		char		*rflag;
		if ( 0 == (rflag = new char[height]) ){
			printf("Memory not enough. \n");
			retcode = 1;
			goto main_exit;
		}
		if ( 0 == (thinimage = sip_CreateImage(width,height,8)) ){
			printf("Memory not enough. \n");
			retcode = 1;
			goto main_exit;
		}
		rect.x = rect.y = 0;
		rect.width  = width;
		rect.height = height;
		for ( i=0 ; i<height ; i++ ){
			sip_cvt1to8((char *)sip_getimgptr(image,i),0,\
				(char *)sip_getimgptr(thinimage,i),width,0,(char)0xff);
			rflag[i] = 1;
		}
		time_start();
		for ( i=0 ; i<1024 ; i++ ){
			if ( 0 >= sip_thin10neib(thinimage,0,rflag) )  break;
		}
		time_get(&time_thinning,0);
	}


//---------------------	Band operation

	time_rlencode = time_hough = time_peakdet = time_lsegcomp = 0;
	time_linecorr = time_areacorr = 0;
	LSEGPOOL.clear();
	for ( xoff=0 ; xoff < (width - (BWID /2)) ; xoff += (BWID /2) ){
		retcode = 0;
		switch ( BandOp.ope(image,thinimage,xoff,&LSEGPOOL, \
			   ARGS.SWmode(),ARGS.SWinfo(),ARGS.SWdirect(),ARGS.SWhw()) ){
		  case 0:	break;
		  case 1:	printf("Too many rectangles.\n");
				retcode = 3;
				break;
		}
		time_rlencode += BandOp.time_rlencode;
		time_hough    += BandOp.time_hough;
		time_peakdet  += BandOp.time_peakdet;
		time_lsegcomp += BandOp.time_lsegcomp;
		time_linecorr += BandOp.time_linecorr;
		time_areacorr += BandOp.time_areacorr;
	}
	if ( ARGS.SWtimeinfo() ){
		printf("Time: Image_thinning     %.3f\n",time_thinning);
		printf("Time: RL_encode          %.3f\n",time_rlencode);
		printf("Time: Hough_transform    %.3f\n",time_hough);
		printf("Time: Peak_detection     %.3f\n",time_peakdet);
		printf("Time: Lseg_compensation  %.3f\n",time_lsegcomp);
		printf("Time: Line_analysis      %.3f\n",time_linecorr);
		printf("Time: Area_analysis      %.3f\n",time_areacorr);
	}


//---------------------	Page operation

	if ( ARGS.SWinfo() )
		printf("Total %d line-segments. (cand-1)\n",LSEGPOOL.counts);

	time_start();  time_rulegrouping = 0;
	if ( ARGS.SWmode() != SW_MODE_ALL ){
		int	rc;
		if ( 0 != PageOp.alloc(LSEGPOOL.counts) ){
			fprintf(stderr,"Group creation error. (alloc)\n");
		}
//		PageOp.mark_group(&LSEGPOOL);
		if ( rc = PageOp.create_groups(&LSEGPOOL,0,0,0) ){
			fprintf(stderr,"Group creation error. (generate) (%d)\n",rc);
		}
//		PageOp.print_groups();
		PageOp.delete_alone(&LSEGPOOL);
		LSEGPOOL.trunc();
	}
	if ( ARGS.SWinfo() )
		printf("Total %d line-segments. (cand-2)\n",LSEGPOOL.counts);

	if ( ARGS.SWmode() == SW_MODE_RULE ){
		int	rc;
		PageOp.clearall();
		if ( 0 != PageOp.alloc(LSEGPOOL.counts) ){
			fprintf(stderr,"Group creation error. (alloc)\n");
		}
		if ( rc = PageOp.create_groups(&LSEGPOOL,0,0,1) ){
			fprintf(stderr,"Group creation error. (generate) (%d)\n",rc);
		}
		PageOp.grpsortline_all(&LSEGPOOL);
//		PageOp.print_groups();
		PageOp.grpsplit2_all(&LSEGPOOL);
		PageOp.combine_lsegs_all(&LSEGPOOL);
	}
	time_get(&time_rulegrouping,0);

	if ( ARGS.SWtimeinfo() ){
		printf("Time: Rule_grouping      %.3f\n",time_rulegrouping);
	}

	if ( ARGS.SWmode() == SW_MODE_ALL )
		retcode = LSEGPOOL.savelist(outfile,width,height);
	else	retcode = PageOp.savegroups(outfile,width,height,&LSEGPOOL);
	switch ( retcode ){
	    case 0:	break;
	    case -1:	fprintf(stderr,"%s : creation error.\n",outfile);
			retcode = 4;
			break;
	    case -2:	fprintf(stderr,"%s : write error.\n",outfile);
			retcode = 5;
			break;
	}

main_exit:
	if ( ARGS.SWthin() )  sip_DestroyImage(thinimage);
	sip_DestroyImage(image);
	return(retcode);
}


