.TH getlseg 1 "Oct. 1998"
.SH NAME
getlseg \- find field\-separators (lines, rules) in document image
.SH SYNOPSIS
.B getlseg
.BR
[\-options]
.BR
.I in_file
|
.I stdin
.BR
.I outfile

.SH DESCRIPTION
getlseg finds linear objects such as field\-separators in a document image,
and produces the data of line segment candidates.
getlseg can find various field\-separators; not only solid lines,
but also dotted lines, broken ones, or those with periodic texture.
getlseg detects the line whose length is more than 1cm and whose
skew is from \-45 to 45degs to the horizontal axis of image.

.PP
getlseg reads a document image of raw, pbm(5) format from
.I in_file.
If the string "stdin" is passed instead of the name of a file,
getlseg reads image from standard input.
The resolution of the input image is always assumed to be 
about 200dpi.

.PP
The data of line segment candidates are written to 
.I outfile,
whose format is the extended 8\-tuple version of d4file(4).
The detail of the format is shown in FILE Section of this manual page.

.PP
The processes in this program are divided into 8 phases as shown 
below. "width" and "height" here denote the width and the height of
the image respectively. (See references for more detail of the 
processes.)

.TP
1.
Cut the original image vertically into "bands" with size 64 x height.
The neighboring bands are set to overlap each other by half 
width of band (=32). For each band, the following process of
phase 2 to 5 is applied.
.TP
2.
Image thinning.
.TP
3.
Line segment detection by Hough Transform.
.TP
4.
Estimation of the black pixel density and auto\-correlation factor
using 1\-dimensional sample.
The line segment candidates are checked if they are the part of field\-separator.
.TP
5.
Estimation of the thickness and auto\-correlation factor
using 2\-dimensional sample.
The line segment candidates are checked again if they are the part of field\-separator.
.TP
6.
Connect line segment candidates across all bands and create groups of 
line segment candidates.
Each group is assumed to be the line/rule candidate.
(option dependent)
.TP
7.
Each line segment candidates is checked if it is connected to any other candidate.
If the candidate is alone, and if it is thick enough, it is deleted.
(option dependent)
.TP
8.
Each group of line segment candidates, which is assumed to be 
the line/rule candidate, is checked if it is too thick for field\-separator.
If so, it is deleted.
(option dependent)

.PP
Phase 6,7,8 are optional. The final phase to be performed is determined
by the command option. (See OPTIONS.)


.SH OPTIONS
.TP
.B \-all
The final process is phase 4.
All of the detected line segment candidates will be written to
the output file.
.TP
.B \-rule
The final process is phase 8.
The line/rule candidates which seem to be really field\-separators will
be written to 
the output file. (This is default.)
.TP
.B \-table
The final process is phase 7.
The line/rule candidates which seem to be field\-separators will
be written to 
the output file.
Even in the case of that the auto\-correlation factor for a 
line/rule candidate at phase 5
is small and that the thickness is too large, the candidates will not 
be deleted if there is another line/rule candidate near them.
This is the difference from the mode of \-rule.
If \-table is specified, lines/rules which are crossing with
vertical lines/rules, such as those in table, will be extracted.
The underlines which text string touch will also be extracted.
However, the number of the false candidates may increases
in the text region.
.TP
.B \-r
Read image rotated by 90degs clockwise.
This option is useful to detect vertical lines.
.TP
.B \-h
Limit the skew angle of segment to \-1.8 to +1.8degs.
This option improves the processing speed of Hough Transform.
Note that this option affects only Hough Transform, and that 
the output file may contain more inclined line segments.



.SH FILE
The output data is written to 
.I outfile.
The format of this file is called "D8".
D8\-formatted file has a header which is defined by the following structure.

.PP
.RS
.ft 3
.nf
.ta .5i 1.3i 3i
struct d8file {
        char    d8_magic[8];    /* = "DATA0008" */
        long    d8_attrb;       /* = 0 */
        long    d8_count;       /* = 0 */
        short   d8_llimit[8];   /* lower  limit of each member */
        short   d8_hlimit[8];   /* higher limit of each member */
        short   d8_spc[40];     /* (filler) */
};
.fi
.ft 1
.RE

.PP
.B d8_magic
field is always filled with string "DATA0008".

.B d8_attrb
holds the attribute code of file. The attribute code is always 0.

.B d8_count
is not used. The value is always zero.

.B d8_llimit
and
.B d8_hlimit
arrays hold lower\- and higher\-limits of 8\-tuple data respectively.

.B d8_spc
array is a dummy field, which is filled with zeros.


.PP
The data of line/rule candidates follow the header.
Each candidate is represented by the following structure.

.PP
.RS
.ft 3
.nf
.ta .3i 1.1i 2.1i
struct lseginfo {
	short	x;	/* x\-coordinate of the center of line segment */
	short	y;	/* y\-coordinate of the center of line segment */
	short	a;	/* skew of line segment */
	short	htpeak;	/* peak value at Hough Transform phase */
	short	corroff;	/* peak position of auto\-correlation factor */
	short	corrpeak;	/* peak value of auto\-correlation factor */
	short	dummy1;
	short	dummy2;
};
.fi
.ft 1
.RE

.PP
The origin of the coordinate axes is placed at the left\-top corner of the input image.
The positive ways of x\- and y\-axis are as shown below.

.nf
+\-\-\-\-\-\-\-> +x
|
|
|
v +y
.fi

.PP
The range of the skew factor "
.B a
" is from \-1024 to +1024.
.B a
/1024 corresponds the tangent of the angle of segment to x\-axis.

.PP
.B htpeak
denotes the peak height in the parameter space after Hough Transformation
and filtering. The range of the value is from 12 to about 64.

.PP
.B corroff
denotes the maximum peak position in the series of
auto\-correlation factor at phase 4.
.B corrpeak
denotes the 
height of the maximum peak. If a line segment candidate seems to be
a solid line because the number of black pixels is close to the width of band, 
.B corroff
is set to \-1.

.PP
If 
.B \-table
or 
.B \-rule 
is specified, each line/rule candidate is terminated by
a terminator.
The terminator is defined as the 8\-tuple data whose members 
are all zero, i.e. { 0,0,0,0, 0,0,0,0 }.
If
.B \-all
is specified, the output file contains no terminator.

.PP
The D8\-formatted file which getlseg produce can be converted to
PostScript file by prlst2ps(1) as shown below.

ex)
.BR 
prlst2ps \-lseg d8_file > tmp.ps


.SH EXIT STATUS
The following exit values are returned:
.PD 0
.TP 10
.B 0
Successful completion.
.TP
.B > 0
An error occurred.
.PD

.SH SEE ALSO
d4file(4), pbm(5), prlst2ps(1)

.SH NOTES
This program is just a SAMPLE, so that there may be some limitations or 
reductions.

.PP
getlseg does not give the true positions of the both ends of
field\-separator because the length of elemental line segment is fixed.

.PP
The processing time of this program is not very short due to
the large computation cost of Hough Transform and 
calculation of auto\-correlation factors.


.SH REFERENCES

Hideaki Goto and Hirotomo Aso, 
"An algorithm for extraction of rules and field\-separators in document image",
Trans. IEICE (D\-II), Vol.J78\-D\-II, No.12, pp.1935\-1939, Dec. 1995.
(In Japanese)

.SH AUTHOR
Copyright (C) 1994\-1998 Hideaki Goto
.\" Permission to use, copy, modify, and distribute this software and
.\" its documentation for any purpose is hereby granted without fee,
.\" provided that (i) the above copyright notice and this permission
.\" notice appear in all copies and in supporting documentation, (ii)
.\" the name of the author, Hideaki Goto, may not be used in any
.\" advertising or otherwise to promote the sale, use or other
.\" dealings in this software without prior written authorization
.\" from the author, (iii) this software may not be used for
.\" commercial products without prior written permission from the
.\" author, and (iv) the notice of the modification is specified in
.\" case of that the modified copies of this software are distributed.
.\"
.\" THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
.\" THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
.\" SOFTWARE.

E\-Mail: hgot@aso.ecei.tohoku.ac.jp
