/*--------------------------------------------------------------
	Peak control functions
		Written   by H.Goto, March 1996
		Rewritten by H.Goto, Sept. 1997
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1996-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>

#include	"siplib.h"
#include	"sgplib.h"

#include	"peakctl.h"


/*------------------------------------------------------
	Peak list control
------------------------------------------------------*/

PEAKS :: PEAKS(){
	peaklist = 0;
	maxpeaks = 0;
	counts = 0;
}


PEAKS :: ~PEAKS(){
	clear();
}


int PEAKS :: incalloc(){
	peakpac*	pktmp;
	int		newsize;
	newsize = maxpeaks + 256;
	if ( 0 == (pktmp = new peakpac[newsize]) )  return(-1);
	memcpy((void *)pktmp,(void *)peaklist,maxpeaks * sizeof(peakpac));
	delete []peaklist;
	peaklist = pktmp;
	maxpeaks = newsize;
	return(0);
}


int PEAKS :: clear(){
	maxpeaks = counts = 0;
	if ( peaklist != 0 ){  delete []peaklist;  peaklist = 0; }
	return(0);
}


int PEAKS :: trunc(){
	int		pkn,newcount;
	peakpac*	pktmp;
	pktmp = peaklist;
	newcount = 0;
	for ( pkn=0 ; pkn<counts ; pkn++ ){
		if ( ! isactive(pkn) )  continue;
		++newcount;
		*pktmp++ = peaklist[pkn];
	}
	counts = newcount;
	return(0);
}


int PEAKS :: setnewpeak(int x,int y,int value){
	if ( counts >= maxpeaks ){
		if ( incalloc() )  return(-1);
	}
	peaklist[counts].stat = PEAKSTAT_ACTIVE;
	peaklist[counts].x = x;
	peaklist[counts].y = y;
	peaklist[counts].value = value;
	peaklist[counts].p = 0;
	peaklist[counts].q = 0;
	++counts;
	return ( counts -1 );
}


int PEAKS :: setnewpeak(int x,int y,int value,int p,int q){
	if ( counts >= maxpeaks ){
		if ( incalloc() )  return(-1);
	}
	peaklist[counts].stat = PEAKSTAT_ACTIVE;
	peaklist[counts].x = x;
	peaklist[counts].y = y;
	peaklist[counts].value = value;
	peaklist[counts].p = p;
	peaklist[counts].q = q;
	++counts;
	return ( counts -1 );
}


int PEAKS :: deletepeak(int pkn){
	if ( pkn >= maxpeaks )  return(-1);
	peaklist[pkn].stat = 0;
	return(0);
}


int PEAKS :: seekto(int id){
	if ( id >= maxpeaks )  return(-1);
	return( out_count = id );
}


int PEAKS :: getnext(int *id,int *x,int *y,int *value){
	for ( ; out_count < counts ; out_count++ ){
		if ( isactive(out_count) ){
			if ( id != 0 )  *id = out_count;
			*x = peaklist[out_count].x;
			*y = peaklist[out_count].y;
			*value = peaklist[out_count].value;
			++out_count;
			return(0);
		}
	}
	return(-1);	/* End of list */
}


int PEAKS :: getnext(int *id,int *x,int *y,int *value,int *p,int *q){
	for ( ; out_count < counts ; out_count++ ){
		if ( isactive(out_count) ){
			if ( id != 0 )  *id = out_count;
			*x = peaklist[out_count].x;
			*y = peaklist[out_count].y;
			*value = peaklist[out_count].value;
			*p = peaklist[out_count].p;
			*q = peaklist[out_count].q;
			++out_count;
			return(0);
		}
	}
	return(-1);	/* End of list */
}


peakpac * PEAKS :: getpeak(int pkn){
	if ( pkn >= maxpeaks )  return(0);
	return( &peaklist[pkn] );
}


/*---------------------------- This function is not included in Class. */
static int peaks_cmp_value(const void *c1,const void *c2){
	return ( ((peakpac *)c2)->value - ((peakpac *)c1)->value );
}
/*---------------------------------------------------------------------*/


int PEAKS :: sortbyvalue(){
	if ( trunc() )  return(-1);
	qsort((char *)peaklist,counts,sizeof(peakpac),peaks_cmp_value);
	return(0);
}




/*------------------------------------------------------
	Peak suppression
------------------------------------------------------*/

static peakbuffer	peakbuf0 = { 0, 0, 0 };


PeakSuppress :: PeakSuppress(){
	len = 0;
	peakbuf = 0;
}


PeakSuppress :: ~PeakSuppress(){
	if ( peakbuf != 0 )  delete []peakbuf;
}


int PeakSuppress :: allocpeakbuf(int len){
	if ( peakbuf != 0 ){
		delete []peakbuf;
		peakbuf = 0;
		len = 0;
	}
	if ( 0 == (peakbuf = new peakbuffer[len]) )  return(-1);
	PeakSuppress::len = len;
	return(0);
}


void PeakSuppress :: clearpeakbuf(){
	int	y;
	for ( y=0 ; y < len ; y++ )  peakbuf[y] = peakbuf0;
}


int PeakSuppress :: setpeaks(int x,peaklisti *pks,int peaks,PEAKS *Pkp,int window){
	int		i,j,retcode;
	int		pos,val;
	int		cs,ce;
	peakbuffer	*peakbuf2;
	if ( 0 == (peakbuf2 = new peakbuffer[len]) )  return(-1);
	for ( i=0 ; i < len ; i++ ){
		peakbuf2[i] = peakbuf0;
	}
	for ( i=0 ; i<peaks ; i++ ){
		pos = pks[i].pos;
		val = pks[i].value;
		peakbuf2[pos].value = val;
		cs = pos - window;  if ( cs < 0 )  cs = 0;
		ce = pos + window;  if ( ce >= len )  ce = len -1;
		for ( j=cs ; j<=ce ; j++ ){
			if ( peakbuf[j].value != 0 ){
				if ( peakbuf[j].value <= val )
					peakbuf[j].rightsup = 1;
				else	peakbuf2[i].leftsup = 1;
			}
		}
	}
	retcode = 0;
	for ( i=0 ; i < len ; i++ ){
		if ( peakbuf[i].value != 0 ){
			if ( (peakbuf[i].leftsup == 0) && (peakbuf[i].rightsup == 0) ){
				if ( 0 > Pkp->setnewpeak(x-1,i,peakbuf[i].value) ){
					retcode = -1;
					break;
				}
			}
		}
	}
	memcpy((void *)peakbuf,(void *)peakbuf2,sizeof(peakbuffer) * len);
	delete []peakbuf2;
	return(retcode);
}




